# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functions for formatting debugging output."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals


def generic_repr(class_instance):
  """Generic debug output for object that lists its property keys and values.

  Use as function:
  x = X()
  generic_repr(x)
  # X(y='hi', z=1)  # ID: 140030868127696

  Use as addition to class:
  class X:
    def __init__(self, y='hi'):
      self.y = y
      self.z = 1

    def __repr__(self):
      return generic_repr(self)
  # X(y='hi', z=1)  # ID: 140030868127696

  Note: Declaring a class by running eval on this repr output will work
  only if all properties are settable in the class's __init__. Nested objects
  may also cause an issue.

  Args:
    class_instance (object): Instance of class to print debug output for.

  Returns:
    Debug output string.
  """
  sorted_attributes = sorted(
      class_instance.__dict__.items(),
      key=lambda key_value_pair: key_value_pair[0])
  return "{class_name}({attributes})  # ID: {id}".format(
      class_name=class_instance.__class__.__name__,
      attributes=", ".join(
          "{}={!r}".format(k, v) for k, v in sorted_attributes),
      id=id(class_instance))



# -*- coding: utf-8 -*- #
# Copyright 2016 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Flags and helpers for the compute instance groups commands."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import enum
import textwrap
from googlecloudsdk.api_lib.compute import managed_instance_groups_utils
from googlecloudsdk.api_lib.compute import utils
from googlecloudsdk.calliope import arg_parsers
from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions
from googlecloudsdk.command_lib.compute import completers as compute_completers
from googlecloudsdk.command_lib.compute import flags
from googlecloudsdk.command_lib.compute import scope as compute_scope
from googlecloudsdk.command_lib.compute.instance_groups.managed import flags as mig_flags
from googlecloudsdk.command_lib.util import completers
import six


STATEFUL_IP_DEFAULT_INTERFACE_NAME = 'nic0'


class RegionalInstanceGroupManagersCompleter(
    compute_completers.ListCommandCompleter):

  def __init__(self, **kwargs):
    super(RegionalInstanceGroupManagersCompleter, self).__init__(
        collection='compute.regionInstanceGroupManagers',
        list_command=('compute instance-groups managed list --uri '
                      '--filter=region:*'),
        **kwargs)


class ZonalInstanceGroupManagersCompleter(
    compute_completers.ListCommandCompleter):

  def __init__(self, **kwargs):
    super(ZonalInstanceGroupManagersCompleter, self).__init__(
        collection='compute.instanceGroupManagers',
        list_command=('compute instance-groups managed list --uri '
                      '--filter=zone:*'),
        **kwargs)


class InstanceGroupManagersCompleter(completers.MultiResourceCompleter):

  def __init__(self, **kwargs):
    super(InstanceGroupManagersCompleter, self).__init__(
        completers=[RegionalInstanceGroupManagersCompleter,
                    ZonalInstanceGroupManagersCompleter],
        **kwargs)


class AutoDeleteFlag(enum.Enum):
  """CLI flag values for `auto-delete' flag."""

  NEVER = 'never'
  ON_PERMANENT_INSTANCE_DELETION = 'on-permanent-instance-deletion'

  def GetAutoDeleteEnumValue(self, base_enum):
    return base_enum(self.name)

  @staticmethod
  def ValidateAutoDeleteFlag(flag_value, flag_name):
    values = [
        auto_delete_flag_value.value
        for auto_delete_flag_value in AutoDeleteFlag
    ]
    if flag_value not in values:
      raise exceptions.InvalidArgumentException(
          parameter_name=flag_name,
          message='Value for [auto-delete] must be [never] or '
          '[on-permanent-instance-deletion], not [{0}]'.format(flag_value))
    return AutoDeleteFlag(flag_value)

  @staticmethod
  def ValidatorWithFlagName(flag_name):
    def Validator(flag_value):
      return AutoDeleteFlag.ValidateAutoDeleteFlag(flag_value, flag_name)
    return Validator


def MakeZonalInstanceGroupArg(plural=False):
  return flags.ResourceArgument(
      resource_name='instance group',
      completer=compute_completers.InstanceGroupsCompleter,
      plural=plural,
      zonal_collection='compute.instanceGroups',
      zone_explanation=flags.ZONE_PROPERTY_EXPLANATION)

MULTISCOPE_INSTANCE_GROUP_ARG = flags.ResourceArgument(
    resource_name='instance group',
    completer=compute_completers.InstanceGroupsCompleter,
    zonal_collection='compute.instanceGroups',
    regional_collection='compute.regionInstanceGroups',
    zone_explanation=flags.ZONE_PROPERTY_EXPLANATION_NO_DEFAULT,
    region_explanation=flags.REGION_PROPERTY_EXPLANATION_NO_DEFAULT)

MULTISCOPE_INSTANCE_GROUP_MANAGER_ARG = flags.ResourceArgument(
    resource_name='managed instance group',
    completer=InstanceGroupManagersCompleter,
    zonal_collection='compute.instanceGroupManagers',
    regional_collection='compute.regionInstanceGroupManagers',
    zone_explanation=flags.ZONE_PROPERTY_EXPLANATION_NO_DEFAULT,
    region_explanation=flags.REGION_PROPERTY_EXPLANATION_NO_DEFAULT)

MULTISCOPE_INSTANCE_GROUP_MANAGERS_ARG = flags.ResourceArgument(
    resource_name='managed instance group',
    plural=True,
    name='names',
    completer=InstanceGroupManagersCompleter,
    zonal_collection='compute.instanceGroupManagers',
    regional_collection='compute.regionInstanceGroupManagers',
    zone_explanation=flags.ZONE_PROPERTY_EXPLANATION_NO_DEFAULT,
    region_explanation=flags.REGION_PROPERTY_EXPLANATION_NO_DEFAULT)


def AddGroupArg(parser):
  parser.add_argument(
      'group',
      help='The name of the instance group.')


def AddNamedPortsArgs(parser):
  """Adds flags for handling named ports."""
  parser.add_argument(
      '--named-ports',
      required=True,
      type=arg_parsers.ArgList(),
      metavar='NAME:PORT',
      help="""\
          The comma-separated list of key:value pairs representing
          the service name and the port that it is running on.

          To clear the list of named ports pass empty list as flag value.
          For example:

            $ {command} example-instance-group --named-ports ""
          """)


def AddScopeArgs(parser, multizonal):
  """Adds flags for group scope."""
  if multizonal:
    scope_parser = parser.add_mutually_exclusive_group()
    flags.AddRegionFlag(
        scope_parser,
        resource_type='instance group',
        operation_type='set named ports for',
        explanation=flags.REGION_PROPERTY_EXPLANATION_NO_DEFAULT)
    flags.AddZoneFlag(
        scope_parser,
        resource_type='instance group',
        operation_type='set named ports for',
        explanation=flags.ZONE_PROPERTY_EXPLANATION_NO_DEFAULT)
  else:
    flags.AddZoneFlag(
        parser,
        resource_type='instance group',
        operation_type='set named ports for')


def AddZonesFlag(parser):
  """Add flags for choosing zones for regional managed instance group."""
  parser.add_argument(
      '--zones',
      metavar='ZONE',
      help="""\
          If this flag is specified a regional managed instance group will be
          created. The managed instance group will be in the same region as
          specified zones and will spread instances in it between specified
          zones.

          All zones must belong to the same region. You may specify --region
          flag but it must be the region to which zones belong. This flag is
          mutually exclusive with --zone flag.""",
      type=arg_parsers.ArgList(min_length=1),
      completer=compute_completers.ZonesCompleter,
      default=[])


def ValidateManagedInstanceGroupScopeArgs(args, resources):
  """Validate arguments specifying scope of the managed instance group."""
  ignored_required_params = {'project': 'fake'}
  if args.zones and args.zone:
    raise exceptions.ConflictingArgumentsException('--zone', '--zones')
  zone_names = []
  for zone in args.zones:
    zone_ref = resources.Parse(
        zone, collection='compute.zones', params=ignored_required_params)
    zone_names.append(zone_ref.Name())

  zone_regions = set([utils.ZoneNameToRegionName(z) for z in zone_names])
  if len(zone_regions) > 1:
    raise exceptions.InvalidArgumentException(
        '--zones', 'All zones must be in the same region.')
  elif len(zone_regions) == 1 and args.region:
    zone_region = zone_regions.pop()
    region_ref = resources.Parse(args.region, collection='compute.regions',
                                 params=ignored_required_params)
    region = region_ref.Name()
    if zone_region != region:
      raise exceptions.InvalidArgumentException(
          '--zones', 'Specified zones not in specified region.')


def ValidateStatefulDisksDict(stateful_disks, flag_name):
  """Validate device-name and auto-delete flags in a stateful disk."""
  device_names = set()
  for stateful_disk in stateful_disks or []:
    if not stateful_disk.get('device-name'):
      raise exceptions.InvalidArgumentException(
          parameter_name=flag_name, message='[device-name] is required')
    if stateful_disk.get('device-name') in device_names:
      raise exceptions.InvalidArgumentException(
          parameter_name=flag_name,
          message='[device-name] `{0}` is not unique in the collection'.format(
              stateful_disk.get('device-name')))
    device_names.add(stateful_disk.get('device-name'))


def ValidateStatefulIPDicts(stateful_ips, flag_name):
  """Validate enabled, interface-name and auto-delete flags in a stateful IP."""
  interface_names = set()
  for stateful_ip in stateful_ips or []:
    # One of: interface-name, enabled is required.
    if (not (stateful_ip.get('interface-name')
             or 'enabled' in stateful_ip)):
      raise exceptions.InvalidArgumentException(
          parameter_name=flag_name,
          message=(
              'one of: [interface-name], [enabled] is required.'))

    interface_name = stateful_ip.get('interface-name',
                                     STATEFUL_IP_DEFAULT_INTERFACE_NAME)

    # Don't accept multiple flags affecting the same interface.
    if interface_name in interface_names:
      raise exceptions.InvalidArgumentException(
          parameter_name=flag_name,
          message=
          '[interface-name] `{0}` is not unique in the collection'.format(
              interface_name))
    interface_names.add(interface_name)


def ValidateManagedInstanceGroupStatefulDisksProperties(args):
  ValidateStatefulDisksDict(args.stateful_disk, '--stateful-disk')


def ValidateManagedInstanceGroupStatefulIPsProperties(args):
  ValidateStatefulIPDicts(args.stateful_internal_ip, '--stateful-internal-ip')
  ValidateStatefulIPDicts(args.stateful_external_ip, '--stateful-external-ip')


def GetInstanceGroupManagerArg(zones_flag=False, region_flag=True):
  """Returns ResourceArgument for working with instance group managers."""
  if zones_flag:
    extra_region_info_about_zones_flag = (
        '\n\nIf you specify `--zones` flag this flag must be unspecified '
        'or specify the region to which the zones you listed belong.'
    )
    region_explanation = (flags.REGION_PROPERTY_EXPLANATION_NO_DEFAULT +
                          extra_region_info_about_zones_flag)
  else:
    region_explanation = flags.REGION_PROPERTY_EXPLANATION_NO_DEFAULT
  if region_flag:
    regional_collection = 'compute.regionInstanceGroupManagers'
  else:
    regional_collection = None
  return flags.ResourceArgument(
      resource_name='managed instance group',
      completer=InstanceGroupManagersCompleter,
      zonal_collection='compute.instanceGroupManagers',
      regional_collection=regional_collection,
      zone_explanation=flags.ZONE_PROPERTY_EXPLANATION_NO_DEFAULT,
      region_explanation=region_explanation)


def CreateGroupReference(client, resources, args):
  resource_arg = GetInstanceGroupManagerArg()
  default_scope = compute_scope.ScopeEnum.ZONE
  scope_lister = flags.GetDefaultScopeLister(client)
  return resource_arg.ResolveAsResource(
      args, resources, default_scope=default_scope,
      scope_lister=scope_lister)


_LIST_INSTANCES_FORMAT = """\
        table(instance.basename():label=NAME,
              instance.scope().segment(0):label=ZONE,
              instanceStatus:label=STATUS,
              instanceHealth[0].detailedHealthState:label=HEALTH_STATE,
              currentAction:label=ACTION,
              version.instanceTemplate.basename():label=INSTANCE_TEMPLATE,
              version.name:label=VERSION_NAME,
              lastAttempt.errors.errors.map().format(
                "Error {0}: {1}", code, message).list(separator=", ")
                :label=LAST_ERROR
        )"""

_LIST_INSTANCES_FORMAT_BETA = """\
        table(instance.basename():label=NAME,
              instance.scope().segment(0):label=ZONE,
              instanceStatus:label=STATUS,
              instanceHealth[0].detailedHealthState:label=HEALTH_STATE,
              currentAction:label=ACTION,
              preservedState():label=PRESERVED_STATE,
              version.instanceTemplate.basename():label=INSTANCE_TEMPLATE,
              version.name:label=VERSION_NAME,
              lastAttempt.errors.errors.map().format(
                "Error {0}: {1}", code, message).list(separator=", ")
                :label=LAST_ERROR
        )"""

_LIST_INSTANCES_FORMAT_ALPHA = """\
        table(instance.basename():label=NAME,
              instance.scope().segment(0):label=ZONE,
              instanceStatus:label=STATUS,
              instanceHealth[0].detailedHealthState:label=HEALTH_STATE,
              currentAction:label=ACTION,
              preservedState():label=PRESERVED_STATE,
              version.instanceTemplate.basename():label=INSTANCE_TEMPLATE,
              version.name:label=VERSION_NAME,
              lastAttempt.errors.errors.map().format(
                "Error {0}: {1}", code, message).list(separator=", ")
                :label=LAST_ERROR
        )"""

_RELEASE_TRACK_TO_LIST_INSTANCES_FORAMT = {
    base.ReleaseTrack.GA: _LIST_INSTANCES_FORMAT,
    base.ReleaseTrack.BETA: _LIST_INSTANCES_FORMAT_BETA,
    base.ReleaseTrack.ALPHA: _LIST_INSTANCES_FORMAT_ALPHA,
}


def _TransformPreservedState(instance):
  """Transform for the PRESERVED_STATE field in the table output.

  PRESERVED_STATE is generated from the fields preservedStateFromPolicy and
  preservedStateFromConfig fields in the managedInstance message.

  Args:
    instance: instance dictionary for transform

  Returns:
    Preserved state status as one of ('POLICY', 'CONFIG', 'POLICY,CONFIG')
  """
  preserved_state_value = ''
  if ('preservedStateFromPolicy' in instance and
      instance['preservedStateFromPolicy']):
    preserved_state_value += 'POLICY,'
  if ('preservedStateFromConfig' in instance and
      instance['preservedStateFromConfig']):
    preserved_state_value += 'CONFIG'
  if preserved_state_value.endswith(','):
    preserved_state_value = preserved_state_value[:-1]
  return preserved_state_value


def AddListInstancesOutputFormat(parser, release_track=base.ReleaseTrack.GA):
  parser.display_info.AddTransforms({
      'preservedState': _TransformPreservedState,
  })
  parser.display_info.AddFormat(
      _RELEASE_TRACK_TO_LIST_INSTANCES_FORAMT[release_track])


# Rename ot HELP_BASE
STATEFUL_DISKS_HELP_BASE = """
      Disks considered stateful by the instance group. Managed instance groups
      preserve and reattach stateful disks on VM autohealing, update, and
      recreate events.
      """

STATEFUL_DISKS_HELP_INSTANCE_CONFIGS = STATEFUL_DISKS_HELP_BASE + """
      You can also attach and preserve disks, not defined in the group's
      instance template, to a given instance.

      The same disk can be attached to more than one instance but only in
      read-only mode.
      """

STATEFUL_DISKS_HELP_INSTANCE_CONFIGS_UPDATE = (
    STATEFUL_DISKS_HELP_INSTANCE_CONFIGS + """
      Use this argument multiple times to update multiple disks.

      If stateful disk with given `device-name` exists in current instance
      configuration, its properties will be replaced by the newly provided ones.
      In other case new stateful disk definition will be added to the instance
      configuration.
      """)

STATEFUL_DISK_DEVICE_NAME_ARG_HELP = """
      *device-name*::: Name under which disk is or will be attached.
      """

STATEFUL_DISK_SOURCE_ARG_HELP = """
      *source*::: Optional argument used to specify the URI of an existing
      persistent disk to attach under specified `device-name`.
      """

STATEFUL_DISK_MODE_ARG_HELP = """
      *mode*::: Specifies the mode of the disk to attach. Supported options are
      `ro` for read-only and `rw` for read-write. If omitted when source is
      specified, `rw` is used as a default. `mode` can only be specified if
      `source` is given.
      """

STATEFUL_DISK_AUTO_DELETE_ARG_HELP = """
      *auto-delete*::: (Optional) Specifies the auto deletion policy of the
      stateful disk. The following options are available:
      - ``never'': (Default) Never delete this disk. Instead, detach the disk
          when its instance is deleted.
      - ``on-permanent-instance-deletion'': Delete the stateful disk when the
          instance that it's attached to is permanently deleted from the group;
          for example, when the instance is deleted manually or when the group
          size is decreased.
      """


STATEFUL_METADATA_HELP = """
      Additional metadata to be made available to the guest operating system
      in addition to the metadata defined in the instance template.

      Stateful metadata may be used to define a key/value pair specific for
      the one given instance to differentiate it from the other instances in
      the managed instance group.

      Stateful metadata key/value pairs are preserved on instance recreation,
      autohealing, updates, and any other lifecycle transitions of the
      instance.

      Stateful metadata have priority over the metadata defined in the
      instance template. This means that stateful metadata that is defined for a
      key that already exists in the instance template overrides the instance
      template value.

      Each metadata entry is a key/value pair separated by an equals sign.
      Metadata keys must be unique and less than 128 bytes in length. Multiple
      entries can be passed to this flag, e.g.,
      ``{argument_name} key-1=value-1,key-2=value-2,key-3=value-3''.
      """

STATEFUL_METADATA_HELP_UPDATE = """
      If stateful metadata with the given key exists in current instance
      configuration, its value will be overridden with the newly provided one.
      If the key does not exist in the current instance configuration, a new
      key/value pair will be added.
      """

STATEFUL_IPS_HELP_BASE = """
      Managed instance groups preserve stateful IPs on VM autohealing, update,
      and recreate events.
      """

STATEFUL_IPS_HELP_TEMPLATE = """
      Use this argument multiple times to update more IPs.

      If a stateful {ip_type} IP with the given interface name already exists in
      the current instance configuration, its properties are replaced by the
      newly provided ones. Otherwise, a new stateful {ip_type} IP definition
      is added to the instance configuration.
      """

STATEFUL_IPS_HELP_INSTANCE_CONFIGS = STATEFUL_IPS_HELP_BASE + """
      You can preserve the IP address that's specified in a network interface
      for a specific managed instance, even if that network interface is not
      defined in the group's instance template.
      """

STATEFUL_IPS_HELP_INSTANCE_CONFIGS_UPDATE = (
    STATEFUL_IPS_HELP_INSTANCE_CONFIGS + """
      Use this argument multiple times to update multiple IPs.

      If a stateful IP with the given network interface name exists in the
      current per-instance configuration, its properties are replaced by
      the newly provided ones. Otherwise, a new stateful IP definition is
      added to the per-instance configuration.
      """)


STATEFUL_IP_ENABLED_ARG_HELP = """
      *enabled*::: Marks the IP address as stateful. The network interface
      named ``nic0'' is assumed by default when ``interface-name'' is not
      specified. This flag can be omitted when ``interface-name'' is provided
      explicitly.
      """


STATEFUL_IP_INTERFACE_NAME_ARG_WITH_ENABLED_HELP = """
      *interface-name*::: Marks the IP address from this network interface as
      stateful. This flag can be omitted when ``enabled'' is provided.
      """


STATEFUL_IP_INTERFACE_NAME_ARG_WITH_ADDRESS_HELP = """
      *interface-name*::: (Optional) Network interface name. If omitted,
      the default network interface named ``nic0'' is assumed.
      """

STATEFUL_IP_ADDRESS_ARG_HELP_INFIX = """
      + Address: URL of a static IP address reservation. For example:
      ``projects/example-project/regions/us-east1/addresses/example-ip-name''.

      + Literal: For example: ``130.211.181.55''.

      If the provided IP address is not yet reserved, the managed instance group
      automatically creates the corresponding IP address reservation. If the
      provided IP address is reserved, the group assigns the reservation to
      the instance.
      """


STATEFUL_IP_ADDRESS_ARG_HELP = """
        *address*::: Static IP address to assign to the instance in one of
        the following formats:
        """ + STATEFUL_IP_ADDRESS_ARG_HELP_INFIX

STATEFUL_IP_ADDRESS_ARG_OPTIONAL_HELP = """
      *address*::: (Optional) Static IP address to assign to the instance in
      one of the following formats:
      """ + STATEFUL_IP_ADDRESS_ARG_HELP_INFIX + """
      The ``address'' flag is optional if an address is already defined in
      the instance's per-instance configuration. Otherwise it is required.

      If omitted, the currently configured address remains unchanged.
      """

STATEFUL_IP_AUTO_DELETE_ARG_HELP = """
      *auto-delete*::: (Optional) Prescribes what should happen to an associated
      static Address resource when a VM instance is permanently deleted.
      Regardless of the value of the delete rule, stateful IP addresses are
      always preserved on instance autohealing, update, and recreation
      operations. The following options are available:
      - ``never'': (Default) Never delete the static IP address. Instead,
          unassign the address when its instance is permanently deleted and
          keep the address reserved.
      - ``on-permanent-instance-deletion'': Delete the static IP
          address reservation when the instance that it's assigned to is
          permanently deleted from the instance group; for example, when the
          instance is deleted manually or when the group size is decreased.
      """


def AddMigCreateStatefulFlags(parser):
  """Adding stateful flags for disks and names to the parser."""
  stateful_disks_help = textwrap.dedent(STATEFUL_DISKS_HELP_BASE + """
      Use this argument multiple times to attach more disks.

      *device-name*::: (Required) Device name of the disk to mark stateful.
      """ + STATEFUL_DISK_AUTO_DELETE_ARG_HELP)
  parser.add_argument(
      '--stateful-disk',
      type=arg_parsers.ArgDict(
          spec={
              'device-name':
                  str,
              'auto-delete': AutoDeleteFlag.ValidatorWithFlagName(
                  '--stateful_disk'),
          }),
      action='append',
      help=stateful_disks_help,
  )


def AddMigCreateStatefulIPsFlags(parser):
  """Adding stateful IPs flags to the parser."""
  stateful_internal_ips_help = textwrap.dedent(
      """
      Internal IPs considered stateful by the instance group. {}
      Use this argument multiple times to make more internal IPs stateful.

      At least one of the following is required:
      {}
      {}

      Additional arguments:
      {}
      """.format(STATEFUL_IPS_HELP_BASE,
                 STATEFUL_IP_ENABLED_ARG_HELP,
                 STATEFUL_IP_INTERFACE_NAME_ARG_WITH_ENABLED_HELP,
                 STATEFUL_IP_AUTO_DELETE_ARG_HELP))
  parser.add_argument(
      '--stateful-internal-ip',
      type=arg_parsers.ArgDict(
          allow_key_only=True,
          spec={
              'enabled': None,
              'interface-name': str,
              'auto-delete': AutoDeleteFlag.ValidatorWithFlagName(
                  '--stateful-internal-ip'),
          }),
      action='append',
      help=stateful_internal_ips_help,
  )

  stateful_external_ips_help = textwrap.dedent(
      """
      External IPs considered stateful by the instance group. {}
      Use this argument multiple times to make more external IPs stateful.

      At least one of the following is required:
      {}
      {}

      Additional arguments:
      {}
      """.format(STATEFUL_IPS_HELP_BASE,
                 STATEFUL_IP_ENABLED_ARG_HELP,
                 STATEFUL_IP_INTERFACE_NAME_ARG_WITH_ENABLED_HELP,
                 STATEFUL_IP_AUTO_DELETE_ARG_HELP))
  parser.add_argument(
      '--stateful-external-ip',
      type=arg_parsers.ArgDict(
          allow_key_only=True,
          spec={
              'enabled': None,
              'interface-name': str,
              'auto-delete': AutoDeleteFlag.ValidatorWithFlagName(
                  '--stateful-external-ip'),
          }),
      action='append',
      help=stateful_external_ips_help,
  )


def _AddMigStatefulInstanceConfigsInstanceArg(parser):
  parser.add_argument(
      '--instance',
      required=True,
      help="""
        URI/name of an existing instance in the managed instance group.
      """)


def AddMigStatefulFlagsForUpdateInstanceConfigs(parser):
  """Add args for per-instance configs update command."""
  _AddMigStatefulInstanceConfigsInstanceArg(parser)

  # Add stateful disk update args
  stateful_disk_argument_name = '--stateful-disk'
  disk_help_text = textwrap.dedent(
      (STATEFUL_DISKS_HELP_INSTANCE_CONFIGS_UPDATE +
       STATEFUL_DISK_DEVICE_NAME_ARG_HELP + STATEFUL_DISK_SOURCE_ARG_HELP +
       STATEFUL_DISK_MODE_ARG_HELP + STATEFUL_DISK_AUTO_DELETE_ARG_HELP))
  parser.add_argument(
      stateful_disk_argument_name,
      type=arg_parsers.ArgDict(
          spec={
              'device-name':
                  str,
              'source':
                  str,
              'mode':
                  str,
              'auto-delete':
                  AutoDeleteFlag.ValidatorWithFlagName(
                      stateful_disk_argument_name)
          }),
      action='append',
      help=disk_help_text,
  )
  # Add remove disk args
  parser.add_argument(
      '--remove-stateful-disks',
      metavar='DEVICE_NAME',
      type=arg_parsers.ArgList(min_length=1),
      help=('Remove stateful configuration for the specified disks from the '
            'instance\'s configuration.'),
  )

  # Add stateful metadata args
  stateful_metadata_argument_name = '--stateful-metadata'
  metadata_help_text = textwrap.dedent(
      (STATEFUL_METADATA_HELP + STATEFUL_METADATA_HELP_UPDATE).format(
          argument_name=stateful_metadata_argument_name))
  parser.add_argument(
      stateful_metadata_argument_name,
      type=arg_parsers.ArgDict(min_length=1),
      default={},
      action=arg_parsers.StoreOnceAction,
      metavar='KEY=VALUE',
      help=textwrap.dedent(metadata_help_text))
  parser.add_argument(
      '--remove-stateful-metadata',
      metavar='KEY',
      type=arg_parsers.ArgList(min_length=1),
      help=('Remove stateful configuration for the specified metadata keys '
            'from the instance\'s configuration.'),
  )


def _AddMigStatefulIPsFlags(parser,
                            ip_argument_name, ip_help_text,
                            remove_ip_argument_name, remove_ip_help_text):
  """Add args for per-instance configs update command."""
  parser.add_argument(
      ip_argument_name,
      type=arg_parsers.ArgDict(
          spec={
              'interface-name':
                  str,
              'address':
                  str,
              'auto-delete':
                  AutoDeleteFlag.ValidatorWithFlagName(ip_argument_name)
          }),
      action='append',
      help=ip_help_text,
  )
  parser.add_argument(
      remove_ip_argument_name,
      metavar='KEY',
      type=arg_parsers.ArgList(min_length=1),
      help=remove_ip_help_text,
  )


def AddMigStatefulIPsFlagsForUpdateInstanceConfigs(parser):
  """Add args for per-instance configs update command."""
  ip_help_text = textwrap.dedent(
      (STATEFUL_IPS_HELP_INSTANCE_CONFIGS_UPDATE +
       STATEFUL_IP_INTERFACE_NAME_ARG_WITH_ADDRESS_HELP +
       STATEFUL_IP_ADDRESS_ARG_OPTIONAL_HELP +
       STATEFUL_IP_AUTO_DELETE_ARG_HELP))
  remove_ip_help_text = """
      List of all stateful IP network interface names to remove from
      the instance's per-instance configuration.
      """
  _AddMigStatefulIPsFlags(parser,
                          '--stateful-internal-ip', ip_help_text,
                          '--remove-stateful-internal-ips', remove_ip_help_text)
  _AddMigStatefulIPsFlags(parser,
                          '--stateful-external-ip', ip_help_text,
                          '--remove-stateful-external-ips', remove_ip_help_text)


def AddMigStatefulFlagsForInstanceConfigs(parser):
  """Adding stateful flags for creating instance configs."""
  _AddMigStatefulInstanceConfigsInstanceArg(parser)

  # Add stateful disk args
  stateful_disk_argument_name = '--stateful-disk'
  stateful_disks_help = textwrap.dedent(
      (STATEFUL_DISKS_HELP_INSTANCE_CONFIGS + """
        Use this argument multiple times to attach and preserve multiple disks.
      """ + STATEFUL_DISK_DEVICE_NAME_ARG_HELP + STATEFUL_DISK_SOURCE_ARG_HELP +
       STATEFUL_DISK_MODE_ARG_HELP + STATEFUL_DISK_AUTO_DELETE_ARG_HELP))
  parser.add_argument(
      stateful_disk_argument_name,
      type=arg_parsers.ArgDict(
          spec={
              'device-name':
                  str,
              'source':
                  str,
              'mode':
                  str,
              'auto-delete':
                  AutoDeleteFlag.ValidatorWithFlagName(
                      stateful_disk_argument_name)
          }),
      action='append',
      help=stateful_disks_help,
  )

  # Add stateful metadata args
  stateful_metadata_argument_name = '--stateful-metadata'
  metadata_help_text = textwrap.dedent(
      STATEFUL_METADATA_HELP.format(
          argument_name=stateful_metadata_argument_name))
  parser.add_argument(
      stateful_metadata_argument_name,
      type=arg_parsers.ArgDict(min_length=1),
      default={},
      action=arg_parsers.StoreOnceAction,
      metavar='KEY=VALUE',
      help=metadata_help_text)


def AddMigStatefulIPsFlagsForInstanceConfigs(parser):
  """Adding stateful IPs flags for creating instance configs."""
  # Add stateful internal IP args
  stateful_ip_help = textwrap.dedent(
      """
      {}
      Use this argument multiple times to attach and preserve multiple IPs.

      {}
      {}
      {}
      """.format(STATEFUL_IPS_HELP_INSTANCE_CONFIGS,
                 STATEFUL_IP_INTERFACE_NAME_ARG_WITH_ADDRESS_HELP,
                 STATEFUL_IP_ADDRESS_ARG_HELP,
                 STATEFUL_IP_AUTO_DELETE_ARG_HELP))

  stateful_internal_ip_argument_name = '--stateful-internal-ip'
  parser.add_argument(
      stateful_internal_ip_argument_name,
      type=arg_parsers.ArgDict(
          spec={
              'interface-name':
                  str,
              'address':
                  str,
              'auto-delete':
                  AutoDeleteFlag.ValidatorWithFlagName(
                      stateful_internal_ip_argument_name)
          }),
      action='append',
      help=stateful_ip_help,
  )

  # Add stateful external IP args
  stateful_external_ip_argument_name = '--stateful-external-ip'
  parser.add_argument(
      stateful_external_ip_argument_name,
      type=arg_parsers.ArgDict(
          spec={
              'interface-name':
                  str,
              'address':
                  str,
              'auto-delete':
                  AutoDeleteFlag.ValidatorWithFlagName(
                      stateful_external_ip_argument_name)
          }),
      action='append',
      help=stateful_ip_help,
  )


def AddCreateInstancesFlags(parser, add_stateful_ips=False):
  """Adding stateful flags for creating and updating instance configs."""
  parser.add_argument(
      '--instance',
      required=True,
      help="""Name of the new instance to create.""")
  parser.add_argument(
      '--stateful-disk',
      type=arg_parsers.ArgDict(
          spec={
              'device-name':
                  str,
              'source':
                  str,
              'mode':
                  str,
              'auto-delete':
                  AutoDeleteFlag.ValidatorWithFlagName('--stateful-disk'),
          }),
      action='append',
      help=textwrap.dedent(STATEFUL_DISKS_HELP_INSTANCE_CONFIGS),
  )
  stateful_metadata_argument_name = '--stateful-metadata'
  parser.add_argument(
      stateful_metadata_argument_name,
      type=arg_parsers.ArgDict(min_length=1),
      default={},
      action=arg_parsers.StoreOnceAction,
      metavar='KEY=VALUE',
      help=textwrap.dedent(
          STATEFUL_METADATA_HELP.format(
              argument_name=stateful_metadata_argument_name)))

  if add_stateful_ips:
    stateful_ips_help_text_template = textwrap.dedent(
        STATEFUL_IPS_HELP_BASE +
        STATEFUL_IPS_HELP_TEMPLATE +
        STATEFUL_IP_INTERFACE_NAME_ARG_WITH_ADDRESS_HELP +
        STATEFUL_IP_ADDRESS_ARG_HELP +
        STATEFUL_IP_AUTO_DELETE_ARG_HELP)

    stateful_internal_ip_flag_name = '--stateful-internal-ip'
    parser.add_argument(
        stateful_internal_ip_flag_name,
        type=arg_parsers.ArgDict(
            spec={
                'interface-name': str,
                'address': str,
                'auto-delete': AutoDeleteFlag.ValidatorWithFlagName(
                    stateful_internal_ip_flag_name)
            }),
        action='append',
        help=stateful_ips_help_text_template.format(ip_type='internal'),
    )

    stateful_external_ip_flag_name = '--stateful-external-ip'
    parser.add_argument(
        stateful_external_ip_flag_name,
        type=arg_parsers.ArgDict(
            spec={
                'interface-name': str,
                'address': str,
                'auto-delete': AutoDeleteFlag.ValidatorWithFlagName(
                    stateful_external_ip_flag_name)
            }),
        action='append',
        help=stateful_ips_help_text_template.format(ip_type='external'),
    )


def AddMigStatefulUpdateInstanceFlag(parser):
  """Add flags for applying updates on PIC change."""
  parser.add_argument(
      '--update-instance',
      default=True,
      action='store_true',
      help="""
          Apply the configuration changes immediately to the instance. If you
          disable this flag, the managed instance group will apply the
          configuration update when you next recreate or update the instance.

          Example: say you have an instance with a disk attached to it and you
          created a stateful configuration for the disk. If you decide to
          delete the stateful configuration for the disk and you provide this
          flag, the group immediately refreshes the instance and removes the
          stateful configuration for the disk. Similarly if you have attached
          a new disk or changed its definition, with this flag the group
          immediately refreshes the instance with the new configuration.""")
  parser.add_argument(
      '--instance-update-minimal-action',
      choices=mig_flags.INSTANCE_ACTION_CHOICES_WITH_NONE,
      default='none',
      help="""
          Perform at least this action on the instance while updating, if
          `--update-instance` is set to `true`.""")


def ValidateMigStatefulDiskFlagForInstanceConfigs(stateful_disks,
                                                  flag_name,
                                                  for_update=False,
                                                  need_disk_source=False):
  """Validates the values of stateful disk flags for instance configs."""
  device_names = set()
  for stateful_disk in stateful_disks or []:
    if not stateful_disk.get('device-name'):
      raise exceptions.InvalidArgumentException(
          parameter_name=flag_name, message='[device-name] is required')

    if stateful_disk.get('device-name') in device_names:
      raise exceptions.InvalidArgumentException(
          parameter_name=flag_name,
          message='[device-name] `{0}` is not unique in the collection'.format(
              stateful_disk.get('device-name')))
    device_names.add(stateful_disk.get('device-name'))

    mode_value = stateful_disk.get('mode')
    if mode_value and mode_value not in ('rw', 'ro'):
      raise exceptions.InvalidArgumentException(
          parameter_name=flag_name,
          message='Value for [mode] must be [rw] or [ro], not [{0}]'.format(
              mode_value))

    if need_disk_source and not stateful_disk.get('source'):
      raise exceptions.InvalidArgumentException(
          parameter_name=flag_name,
          message='[source] is required for all stateful disks')

    if not for_update and mode_value and not stateful_disk.get('source'):
      raise exceptions.InvalidArgumentException(
          parameter_name=flag_name,
          message='[mode] can be set then and only then when [source] is given')


def ValidateMigStatefulIpFlagForInstanceConfigs(flag_name, stateful_ips,
                                                current_addresses):
  """Validates the values of stateful IP flags for instance configs."""
  interface_names = set()
  for stateful_ip in (stateful_ips or []):
    interface_name = stateful_ip.get('interface-name',
                                     STATEFUL_IP_DEFAULT_INTERFACE_NAME)
    if not ('address' in stateful_ip
            or interface_name in current_addresses):
      raise exceptions.InvalidArgumentException(
          parameter_name=flag_name, message='[address] is required')

    if interface_name in interface_names:
      raise exceptions.InvalidArgumentException(
          parameter_name=flag_name,
          message='[interface-name] `{0}` is not unique in the collection'
          .format(interface_name))
    interface_names.add(interface_name)


def ValidateMigStatefulDisksRemovalFlagForInstanceConfigs(disks_to_remove,
                                                          disks_to_update):
  remove_stateful_disks_set = set(disks_to_remove or [])
  for stateful_disk_to_update in disks_to_update or []:
    if stateful_disk_to_update.get('device-name') in remove_stateful_disks_set:
      raise exceptions.InvalidArgumentException(
          parameter_name='--remove-stateful-disks',
          message=('the same [device-name] `{0}` cannot be updated and'
                   ' removed in one command call'.format(
                       stateful_disk_to_update.get('device-name'))))


def ValidateMigStatefulMetadataRemovalFlagForInstanceConfigs(entries_to_remove,
                                                             entries_to_update):
  remove_stateful_metadata_set = set(entries_to_remove or [])
  update_stateful_metadata_set = set(entries_to_update.keys())
  keys_intersection = remove_stateful_metadata_set.intersection(
      update_stateful_metadata_set)
  if keys_intersection:
    raise exceptions.InvalidArgumentException(
        parameter_name='--remove-stateful-metadata',
        message=('the same metadata key(s) `{0}` cannot be updated and'
                 ' removed in one command call'.format(
                     ', '.join(keys_intersection))))


def ValidateMigStatefulIpsRemovalFlagForInstanceConfigs(flag_name,
                                                        ips_to_remove,
                                                        ips_to_update):
  remove_ips_set = set(ips_to_remove or [])
  for ip_to_update in ips_to_update or []:
    if ip_to_update.get('interface-name') in remove_ips_set:
      raise exceptions.InvalidArgumentException(
          parameter_name=flag_name,
          message=('the same [interface-name] `{0}` cannot be updated and'
                   ' removed in one command call'.format(
                       ip_to_update.get('interface-name'))))


def ValidateMigStatefulFlagsForInstanceConfigs(args,
                                               for_update=False,
                                               need_disk_source=False):
  """Validates the values of stateful flags for instance configs."""
  ValidateMigStatefulDiskFlagForInstanceConfigs(args.stateful_disk,
                                                '--stateful-disk',
                                                for_update, need_disk_source)

  if for_update:
    ValidateMigStatefulDisksRemovalFlagForInstanceConfigs(
        disks_to_remove=args.remove_stateful_disks,
        disks_to_update=args.stateful_disk)
    ValidateMigStatefulMetadataRemovalFlagForInstanceConfigs(
        entries_to_remove=args.remove_stateful_metadata,
        entries_to_update=args.stateful_metadata)


def ValidateMigStatefulIPFlagsForInstanceConfigs(args,
                                                 current_internal_addresses,
                                                 current_external_addresses,
                                                 for_update=False):
  """Validates the values of stateful flags for instance configs, with IPs."""
  ValidateMigStatefulIpFlagForInstanceConfigs('--stateful-internal-ip',
                                              args.stateful_internal_ip,
                                              current_internal_addresses)
  ValidateMigStatefulIpFlagForInstanceConfigs('--stateful-external-ip',
                                              args.stateful_external_ip,
                                              current_external_addresses)
  if for_update:
    ValidateMigStatefulIpsRemovalFlagForInstanceConfigs(
        flag_name='--remove-stateful-internal-ips',
        ips_to_remove=args.remove_stateful_internal_ips,
        ips_to_update=args.stateful_internal_ip)
    ValidateMigStatefulIpsRemovalFlagForInstanceConfigs(
        flag_name='--remove-stateful-external-ips',
        ips_to_remove=args.remove_stateful_external_ips,
        ips_to_update=args.stateful_external_ip)


def AddDescriptionFlag(parser, for_update=False):
  """Add --description to the parser."""
  parser.add_argument(
      '--description',
      help='An optional description for this group.' +
      (' To clear the description, set the value to an empty string.'
       if for_update
       else ''))


def AddMigUpdateStatefulFlags(parser):
  """Add --stateful-disk and --remove-stateful-disks to the parser."""
  stateful_disks_help = textwrap.dedent(STATEFUL_DISKS_HELP_BASE + """
      Use this argument multiple times to update more disks.

      If a stateful disk with the given device name already exists in the
      current instance configuration, its properties will be replaced by the
      newly provided ones. Otherwise, a new stateful disk definition will be
      added to the instance configuration.

      *device-name*::: (Required) Device name of the disk to mark stateful.
      """ + STATEFUL_DISK_AUTO_DELETE_ARG_HELP)
  stateful_disk_flag_name = '--stateful-disk'
  parser.add_argument(
      stateful_disk_flag_name,
      type=arg_parsers.ArgDict(
          spec={
              'device-name':
                  str,
              'auto-delete':
                  AutoDeleteFlag.ValidatorWithFlagName(stateful_disk_flag_name)
          }),
      action='append',
      help=stateful_disks_help,
  )
  parser.add_argument(
      '--remove-stateful-disks',
      metavar='DEVICE_NAME',
      type=arg_parsers.ArgList(min_length=1),
      help='Remove stateful configuration for the specified disks.',
  )


def AddMigUpdateStatefulFlagsIPs(parser):
  """Add stateful IPs flags to the parser."""
  stateful_ips_help_text_template = textwrap.dedent(
      STATEFUL_IPS_HELP_BASE +
      STATEFUL_IPS_HELP_TEMPLATE +
      """
      At least one of the following is required:
      {}
      {}

      Additional arguments:
      {}
      """.format(STATEFUL_IP_ENABLED_ARG_HELP,
                 STATEFUL_IP_INTERFACE_NAME_ARG_WITH_ENABLED_HELP,
                 STATEFUL_IP_AUTO_DELETE_ARG_HELP))

  stateful_internal_ip_flag_name = '--stateful-internal-ip'
  parser.add_argument(
      stateful_internal_ip_flag_name,
      type=arg_parsers.ArgDict(
          allow_key_only=True,
          spec={
              'enabled': None,
              'interface-name': str,
              'auto-delete': AutoDeleteFlag.ValidatorWithFlagName(
                  stateful_internal_ip_flag_name)
          }),
      action='append',
      help=stateful_ips_help_text_template.format(ip_type='internal'),
  )

  stateful_external_ip_flag_name = '--stateful-external-ip'
  parser.add_argument(
      stateful_external_ip_flag_name,
      type=arg_parsers.ArgDict(
          allow_key_only=True,
          spec={
              'enabled': None,
              'interface-name': str,
              'auto-delete': AutoDeleteFlag.ValidatorWithFlagName(
                  stateful_external_ip_flag_name)
          }),
      action='append',
      help=stateful_ips_help_text_template.format(ip_type='external'),
  )

  remove_stateful_ips_help_text_template = """
      Remove stateful configuration for the specified interfaces for
      {ip_type} IPs.
      """
  parser.add_argument(
      '--remove-stateful-internal-ips',
      metavar='INTERFACE_NAME',
      type=arg_parsers.ArgList(min_length=1),
      help=remove_stateful_ips_help_text_template.format(ip_type='internal'),
  )
  parser.add_argument(
      '--remove-stateful-external-ips',
      metavar='INTERFACE_NAME',
      type=arg_parsers.ArgList(min_length=1),
      help=remove_stateful_ips_help_text_template.format(ip_type='external'),
  )


def ValidateUpdateStatefulPolicyParams(args, current_stateful_policy):
  """Check stateful properties of update request."""
  current_device_names = set(
      managed_instance_groups_utils.GetDeviceNamesFromStatefulPolicy(
          current_stateful_policy))
  update_disk_names = []
  if args.stateful_disk:
    ValidateStatefulDisksDict(args.stateful_disk, '--stateful-disk')
    update_disk_names = [
        stateful_disk.get('device-name') for stateful_disk in args.stateful_disk
    ]
  if args.remove_stateful_disks:
    if any(
        args.remove_stateful_disks.count(x) > 1
        for x in args.remove_stateful_disks):
      raise exceptions.InvalidArgumentException(
          parameter_name='update',
          message=(
              'When removing device names from Stateful Policy, please provide '
              'each name exactly once.'))

  update_set = set(update_disk_names)
  remove_set = set(args.remove_stateful_disks or [])
  intersection = update_set.intersection(remove_set)

  if intersection:
    raise exceptions.InvalidArgumentException(
        parameter_name='update',
        message=
        ('You cannot simultaneously add and remove the same device names {} to '
         'Stateful Policy.'.format(six.text_type(intersection))))
  not_current_device_names = remove_set - current_device_names
  if not_current_device_names:
    raise exceptions.InvalidArgumentException(
        parameter_name='update',
        message=('Disks [{}] are not currently set as stateful, '
                 'so they cannot be removed from Stateful Policy.'.format(
                     six.text_type(not_current_device_names))))


def _ValidateUpdateStatefulPolicyParamsWithIPsCommon(current_interface_names,
                                                     update_flag_name,
                                                     remove_flag_name,
                                                     update_ips,
                                                     remove_ips,
                                                     ip_type_name):
  """Check stateful properties of update request."""
  update_interface_names = []
  if update_ips:
    ValidateStatefulIPDicts(update_ips, update_flag_name)
    for stateful_ip in update_ips:
      update_interface_names.append(
          stateful_ip.get('interface-name', STATEFUL_IP_DEFAULT_INTERFACE_NAME))

  remove_interface_names = remove_ips or []

  if any(
      remove_interface_names.count(x) > 1
      for x in remove_interface_names):
    raise exceptions.InvalidArgumentException(
        parameter_name='update',
        message=(
            'When removing stateful {} IPs from Stateful Policy, please '
            'provide each network interface name exactly once.'.format(
                ip_type_name)))

  update_set = set(update_interface_names)
  remove_set = set(remove_interface_names)
  intersection = update_set.intersection(remove_set)

  if intersection:
    raise exceptions.InvalidArgumentException(
        parameter_name='update',
        message=
        ('You cannot simultaneously add and remove the same interface {} to '
         'stateful {} IPs in Stateful Policy.'.format(
             six.text_type(intersection), ip_type_name)))
  not_current_interface_names = remove_set - current_interface_names
  if not_current_interface_names:
    raise exceptions.InvalidArgumentException(
        parameter_name='update',
        message=('Interfaces [{}] are not currently set as stateful {} IPs, '
                 'so they cannot be removed from Stateful Policy.'.format(
                     six.text_type(not_current_interface_names), ip_type_name)))


def _ValidateUpdateStatefulPolicyParamsWithInternalIPs(args,
                                                       current_stateful_policy):
  """Check stateful internal IPs properties of update request."""
  current_interface_names = set(
      managed_instance_groups_utils
      .GetInterfaceNamesFromStatefulPolicyForInternalIPs(
          current_stateful_policy))
  _ValidateUpdateStatefulPolicyParamsWithIPsCommon(
      current_interface_names,
      '--stateful-internal-ip', '--remove-stateful-internal-ips',
      args.stateful_internal_ip, args.remove_stateful_internal_ips, 'internal')


def _ValidateUpdateStatefulPolicyParamsWithExternalIPs(args,
                                                       current_stateful_policy):
  """Check stateful external IPs properties of update request."""
  current_interface_names = set(
      managed_instance_groups_utils
      .GetInterfaceNamesFromStatefulPolicyForExternalIPs(
          current_stateful_policy))
  _ValidateUpdateStatefulPolicyParamsWithIPsCommon(
      current_interface_names,
      '--stateful-external-ip', '--remove-stateful-external-ips',
      args.stateful_external_ip, args.remove_stateful_external_ips, 'external')


def ValidateUpdateStatefulPolicyParamsWithIPs(args, current_stateful_policy):
  """Check stateful properties of update request."""
  _ValidateUpdateStatefulPolicyParamsWithInternalIPs(args,
                                                     current_stateful_policy)
  _ValidateUpdateStatefulPolicyParamsWithExternalIPs(args,
                                                     current_stateful_policy)



# -*- coding: utf-8 -*- #
# Copyright 2018 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Useful commands for interacting with the Cloud Firestore Operations API."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from apitools.base.py import list_pager
from googlecloudsdk.api_lib.util import apis
from googlecloudsdk.api_lib.util import waiter
from googlecloudsdk.core import resources

OPERATIONS_API_VERSION = 'v1'
DEFAULT_PAGE_SIZE = 100


def GetClient():
  """Returns the Cloud Firestore client for operations."""
  return apis.GetClientInstance('firestore', OPERATIONS_API_VERSION)


def GetService():
  """Returns the service for interacting with the Operations service."""
  return GetClient().projects_databases_operations


def GetMessages():
  """Import and return the appropriate operations messages module."""
  return apis.GetMessagesModule('firestore', OPERATIONS_API_VERSION)


def ListOperations(project, database, limit=None, operation_filter=None):
  """Lists all of the Firestore operations.

  Args:
    project: the project to list operations for, a string.
    database: the database to list operations for, a string. Defaults to the
      default database.
    limit: the maximum number of operations to return, an integer. Defaults to
      positive infinity if unset.
    operation_filter: a filter to apply to operations, a string.

  Returns:
    a generator of Operations.
  """
  list_request = GetMessages().FirestoreProjectsDatabasesOperationsListRequest(
      filter=operation_filter,
      name='projects/{0}/databases/{1}'.format(project, database))
  batch_size = min(limit, DEFAULT_PAGE_SIZE) if limit else DEFAULT_PAGE_SIZE
  return list_pager.YieldFromList(
      GetService(),
      list_request,
      limit=limit,
      batch_size=batch_size,
      field='operations',
      batch_size_attribute='pageSize')


def GetOperation(name):
  """Returns the google.longrunning.Operation with the given name."""
  return GetService().Get(
      GetMessages().FirestoreProjectsDatabasesOperationsGetRequest(name=name))


def CancelOperation(name):
  """Cancels the Operation with the given name."""
  return GetService().Cancel(
      GetMessages().FirestoreProjectsDatabasesOperationsCancelRequest(
          name=name))


def DeleteOperation(name):
  """Deletes the Operation with the given name."""
  return GetService().Delete(
      GetMessages().FirestoreProjectsDatabasesOperationsDeleteRequest(
          name=name))


def WaitForOperation(operation):
  """Waits for the given Operation to complete."""
  operation_ref = resources.REGISTRY.Parse(
      operation.name,
      collection='firestore.projects.databases.operations',
      api_version=OPERATIONS_API_VERSION)
  poller = waiter.CloudOperationPollerNoResources(GetService(),
                                                  lambda x: x.RelativeName())
  return waiter.WaitFor(
      poller, operation_ref,
      'Waiting for [{0}] to finish'.format(operation_ref.RelativeName()))




# -*- coding: utf-8 -*- #
# Copyright 2020 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for querying custom jobs in AI Platform."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from apitools.base.py import list_pager
from googlecloudsdk.api_lib.util import apis
from googlecloudsdk.command_lib.ai import constants
from googlecloudsdk.command_lib.export import util as export_util
from googlecloudsdk.core.console import console_io


class CustomJobsClient(object):
  """Client used for interacting with CustomJob endpoint."""

  def __init__(self, version=constants.GA_VERSION):
    client = apis.GetClientInstance(constants.AI_PLATFORM_API_NAME,
                                    constants.AI_PLATFORM_API_VERSION[version])
    self._messages = client.MESSAGES_MODULE
    self._version = version
    self._service = client.projects_locations_customJobs
    self._message_prefix = constants.AI_PLATFORM_MESSAGE_PREFIX[version]

  def GetMessage(self, message_name):
    """Returns the API message class by name."""

    return getattr(
        self._messages,
        '{prefix}{name}'.format(prefix=self._message_prefix,
                                name=message_name), None)

  def CustomJobMessage(self):
    """Retures the CustomJob resource message."""

    return self.GetMessage('CustomJob')

  def Create(self,
             parent,
             job_spec,
             display_name=None,
             kms_key_name=None,
             labels=None):
    """Constructs a request and sends it to the endpoint to create a custom job instance.

    Args:
      parent: str, The project resource path of the custom job to create.
      job_spec: The CustomJobSpec message instance for the job creation request.
      display_name: str, The display name of the custom job to create.
      kms_key_name: A customer-managed encryption key to use for the custom job.
      labels: LabelValues, map-like user-defined metadata to organize the custom
        jobs.

    Returns:
      A CustomJob message instance created.
    """
    custom_job = self.CustomJobMessage()(
        displayName=display_name, jobSpec=job_spec)

    if kms_key_name is not None:
      custom_job.encryptionSpec = self.GetMessage('EncryptionSpec')(
          kmsKeyName=kms_key_name)

    if labels:
      custom_job.labels = labels

    if self._version == constants.ALPHA_VERSION:
      return self._service.Create(
          self._messages.AiplatformProjectsLocationsCustomJobsCreateRequest(
              parent=parent, googleCloudAiplatformV1alpha1CustomJob=custom_job))
    elif self._version == constants.BETA_VERSION:
      return self._service.Create(
          self._messages.AiplatformProjectsLocationsCustomJobsCreateRequest(
              parent=parent, googleCloudAiplatformV1beta1CustomJob=custom_job))
    else:
      return self._service.Create(
          self._messages.AiplatformProjectsLocationsCustomJobsCreateRequest(
              parent=parent, googleCloudAiplatformV1CustomJob=custom_job))

  def List(self, limit=None, region=None):
    return list_pager.YieldFromList(
        self._service,
        self._messages.AiplatformProjectsLocationsCustomJobsListRequest(
            parent=region),
        field='customJobs',
        batch_size_attribute='pageSize',
        limit=limit)

  def Get(self, name):
    request = self._messages.AiplatformProjectsLocationsCustomJobsGetRequest(
        name=name)
    return self._service.Get(request)

  def Cancel(self, name):
    request = self._messages.AiplatformProjectsLocationsCustomJobsCancelRequest(
        name=name)
    return self._service.Cancel(request)

  def CheckJobComplete(self, name):
    """Returns a function to decide if log fetcher should continue polling.

    Args:
      name: String id of job.

    Returns:
      A one-argument function decides if log fetcher should continue.
    """
    request = self._messages.AiplatformProjectsLocationsCustomJobsGetRequest(
        name=name)
    response = self._service.Get(request)

    def ShouldContinue(periods_without_logs):
      if periods_without_logs <= 1:
        return True
      return response.endTime is None

    return ShouldContinue

  def ImportResourceMessage(self, yaml_file, message_name):
    """Import a messages class instance typed by name from a YAML file."""
    data = console_io.ReadFromFileOrStdin(yaml_file, binary=False)
    message_type = self.GetMessage(message_name)
    return export_util.Import(message_type=message_type, stream=data)



# -*- coding: utf-8 -*- #
# Copyright 2021 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Client for interaction with LAKE API CRUD DATAPLEX."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.api_lib.dataplex import util as dataplex_api
from googlecloudsdk.api_lib.storage import storage_api
from googlecloudsdk.command_lib.iam import iam_util


def SetIamPolicy(lake_ref, policy):
  """Set Iam Policy request."""
  set_iam_policy_req = dataplex_api.GetMessageModule(
  ).DataplexProjectsLocationsLakesSetIamPolicyRequest(
      resource=lake_ref.RelativeName(),
      googleIamV1SetIamPolicyRequest=dataplex_api.GetMessageModule()
      .GoogleIamV1SetIamPolicyRequest(policy=policy))
  return dataplex_api.GetClientInstance().projects_locations_lakes.SetIamPolicy(
      set_iam_policy_req)


def GetIamPolicy(lake_ref):
  """Get Iam Policy request."""
  get_iam_policy_req = dataplex_api.GetMessageModule(
  ).DataplexProjectsLocationsLakesGetIamPolicyRequest(
      resource=lake_ref.RelativeName())
  return dataplex_api.GetClientInstance().projects_locations_lakes.GetIamPolicy(
      get_iam_policy_req)


def AddIamPolicyBinding(lake_ref, member, role):
  """Add iam policy binding request."""
  policy = GetIamPolicy(lake_ref)
  iam_util.AddBindingToIamPolicy(
      dataplex_api.GetMessageModule().GoogleIamV1Binding, policy, member, role)
  return SetIamPolicy(lake_ref, policy)


def RemoveIamPolicyBinding(lake_ref, member, role):
  """Remove iam policy binding request."""
  policy = GetIamPolicy(lake_ref)
  iam_util.RemoveBindingFromIamPolicy(policy, member, role)
  return SetIamPolicy(lake_ref, policy)


def AddServiceAccountToDatasetPolicy(access_message_type, dataset_policy,
                                     member, role):
  """Add service account to dataset."""
  for entry in dataset_policy.access:
    if entry.role == role and member in entry.userByEmail:
      return False
  dataset_policy.access.append(
      access_message_type(userByEmail=member, role='{0}'.format(role)))
  return True


def SetIamPolicyFromFile(lake_ref, policy_file):
  """Set iam policy binding request from file."""
  policy = iam_util.ParsePolicyFile(
      policy_file,
      dataplex_api.GetMessageModule().GoogleIamV1Policy)
  return SetIamPolicy(lake_ref, policy)


def RemoveServiceAccountFromBucketPolicy(bucket_ref, member, role):
  """Deauthorize Account for Buckets."""
  policy = storage_api.StorageClient().GetIamPolicy(bucket_ref)
  iam_util.RemoveBindingFromIamPolicy(policy, member, role)
  return storage_api.StorageClient().SetIamPolicy(bucket_ref, policy)


def RemoveServiceAccountFromDatasetPolicy(dataset_policy, member, role):
  """Deauthorize Account for Dataset."""
  for entry in dataset_policy.access:
    if entry.role == role and member in entry.userByEmail:
      dataset_policy.access.remove(entry)
      return True
  return False


def GenerateUpdateMask(args):
  """Create Update Mask for Lakes."""
  update_mask = []
  if args.IsSpecified('description'):
    update_mask.append('description')
  if args.IsSpecified('display_name'):
    update_mask.append('displayName')
  if args.IsSpecified('labels'):
    update_mask.append('labels')
  if args.IsSpecified('metastore_service'):
    update_mask.append('metastore.service')
  return update_mask


def WaitForOperation(operation):
  """Waits for the given google.longrunning.Operation to complete."""
  return dataplex_api.WaitForOperation(
      operation,
      dataplex_api.GetClientInstance().projects_locations_lakes)



# -*- coding: utf-8 -*- #
# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Extra utility functions for OS Config Troubleshooter."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import logging
import traceback

from googlecloudsdk.calliope import base
from googlecloudsdk.core import log
from googlecloudsdk.core import properties


def GetCommandString(args, release_track):
  """Gets the command string for the user to invoke.

  Args:
    args: the command args. These should be arguments put after 'gcloud'.
    release_track: The release track.

  Returns:
    The command string, formatted as '$ gcloud [args]'.
  """
  if release_track != base.ReleaseTrack.GA:
    args.insert(0, release_track.prefix)
  return ' '.join(['$', 'gcloud'] + args)


def UnknownMessage(exception):
  debug_verbosity = log.GetVerbosity() == logging.DEBUG
  message = (
      'Unknown\n'
      'The troubleshooter encountered ' + type(exception).__name__ + ' while '
      'checking your instance.'
  )
  if debug_verbosity:
    message += '\n{}\n'.format(traceback.format_exc(limit=0))
  return message


def GetProject(client, project):
  return client.MakeRequests(
      [(client.apitools_client.projects, 'Get',
        client.messages.ComputeProjectsGetRequest(
            project=project or
            properties.VALUES.core.project.Get(required=True),))])[0]


def GetInstance(client, instance_ref):
  request = client.messages.ComputeInstancesGetRequest(
      **instance_ref.AsDict())
  return client.MakeRequests([
      (client.apitools_client.instances, 'Get', request)])[0]


class Response:
  """Represents a response returned by each of the prerequisite checks."""

  def __init__(self, continue_flag, response_message):
    self.continue_flag = continue_flag
    self.response_message = response_message



# -*- coding: utf-8 -*- #
# Copyright 2022 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""VMware Engine network policy client."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from apitools.base.py import list_pager
from googlecloudsdk.api_lib.vmware import util
from googlecloudsdk.api_lib.vmware.networks import NetworksClient
from googlecloudsdk.command_lib.vmware.networks import util as networks_util


class NetworkPoliciesClient(util.VmwareClientBase):
  """VMware Engine network policy client."""

  def __init__(self):
    super(NetworkPoliciesClient, self).__init__()
    self.service = self.client.projects_locations_networkPolicies
    self.networks_client = NetworksClient()

  def Get(self, resource):
    request = self.messages.VmwareengineProjectsLocationsNetworkPoliciesGetRequest(
        name=resource.RelativeName())
    response = self.service.Get(request)
    return response

  def Create(self,
             resource,
             vmware_engine_network_id=None,
             description=None,
             edge_services_cidr=None,
             internet_access=None,
             external_ip_access=None):
    parent = resource.Parent().RelativeName()
    project = resource.Parent().Parent().Name()
    network_policy_id = resource.Name()
    network_policy = self.messages.NetworkPolicy(description=description)
    internet_access_obj = self.messages.NetworkService(enabled=internet_access)
    external_ip_access_obj = self.messages.NetworkService(
        enabled=external_ip_access)
    if vmware_engine_network_id is not None:
      ven = self.networks_client.GetByID(project, vmware_engine_network_id)
      network_policy.vmwareEngineNetwork = ven.name
    network_policy.edgeServicesCidr = edge_services_cidr
    network_policy.internetAccess = internet_access_obj
    network_policy.externalIp = external_ip_access_obj
    request = self.messages.VmwareengineProjectsLocationsNetworkPoliciesCreateRequest(
        parent=parent,
        networkPolicy=network_policy,
        networkPolicyId=network_policy_id,
        requestId=networks_util.GetUniqueId())
    return self.service.Create(request)

  def Update(self,
             resource,
             description=None,
             edge_services_cidr=None,
             internet_access=None,
             external_ip_access=None):
    network_policy = self.Get(resource)
    update_mask = []
    if description is not None:
      network_policy.description = description
      update_mask.append('description')
    if edge_services_cidr is not None:
      network_policy.edgeServicesCidr = edge_services_cidr
      update_mask.append('edge_services_cidr')
    if internet_access is not None:
      internet_access_obj = self.messages.NetworkService(
          enabled=internet_access)
      network_policy.internetAccess = internet_access_obj
      update_mask.append('internet_access.enabled')
    if external_ip_access is not None:
      external_ip_access_obj = self.messages.NetworkService(
          enabled=external_ip_access)
      network_policy.externalIp = external_ip_access_obj
      update_mask.append('external_ip.enabled')
    request = self.messages.VmwareengineProjectsLocationsNetworkPoliciesPatchRequest(
        networkPolicy=network_policy,
        name=resource.RelativeName(),
        updateMask=','.join(update_mask),
        requestId=networks_util.GetUniqueId())
    return self.service.Patch(request)

  def Delete(self, resource):
    return self.service.Delete(
        self.messages
        .VmwareengineProjectsLocationsNetworkPoliciesDeleteRequest(
            name=resource.RelativeName(),
            requestId=networks_util.GetUniqueId()))

  def List(self,
           location_resource,
           filter_expression=None,
           limit=None,
           page_size=None):
    location = location_resource.RelativeName()
    request = self.messages.VmwareengineProjectsLocationsNetworkPoliciesListRequest(
        parent=location, filter=filter_expression)
    if page_size:
      request.page_size = page_size
    return list_pager.YieldFromList(
        self.service,
        request,
        limit=limit,
        batch_size_attribute='pageSize',
        batch_size=page_size,
        field='networkPolicies')



# -*- coding: utf-8 -*- #
# Copyright 2016 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utilities for the cloudbuild API."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import enum
import re

from apitools.base.protorpclite import messages as proto_messages
from apitools.base.py import encoding as apitools_encoding
from googlecloudsdk.api_lib.cloudbuild import cloudbuild_exceptions
from googlecloudsdk.api_lib.util import apis
from googlecloudsdk.calliope import base
from googlecloudsdk.core import yaml
from googlecloudsdk.core.resource import resource_property
from googlecloudsdk.core.util import files

import six

_API_NAME = 'cloudbuild'
_GA_API_VERSION = 'v1'
_BETA_API_VERSION = 'v1beta1'

RELEASE_TRACK_TO_API_VERSION = {
    base.ReleaseTrack.GA: _GA_API_VERSION,
    base.ReleaseTrack.BETA: _GA_API_VERSION,
    base.ReleaseTrack.ALPHA: _GA_API_VERSION,
}

WORKERPOOL_NAME_MATCHER = r'projects/.*/locations/.*/workerPools/.*'
WORKERPOOL_NAME_SELECTOR = r'projects/.*/locations/.*/workerPools/(.*)'
WORKERPOOL_REGION_SELECTOR = r'projects/.*/locations/(.*)/workerPools/.*'

# Default for optionally-regional requests when the user does not specify.
DEFAULT_REGION = 'global'

BYTES_IN_ONE_GB = 2**30


class WorkerpoolTypes(enum.Enum):
  UNKNOWN = 0
  PRIVATE = 1
  HYBRID = 2


def GetMessagesModule(release_track=base.ReleaseTrack.GA):
  """Returns the messages module for Cloud Build.

  Args:
    release_track: The desired value of the enum
      googlecloudsdk.calliope.base.ReleaseTrack.

  Returns:
    Module containing the definitions of messages for Cloud Build.
  """
  return apis.GetMessagesModule(_API_NAME,
                                RELEASE_TRACK_TO_API_VERSION[release_track])


def GetClientClass(release_track=base.ReleaseTrack.GA):
  """Returns the client class for Cloud Build.

  Args:
    release_track: The desired value of the enum
      googlecloudsdk.calliope.base.ReleaseTrack.

  Returns:
    base_api.BaseApiClient, Client class for Cloud Build.
  """
  return apis.GetClientClass(_API_NAME,
                             RELEASE_TRACK_TO_API_VERSION[release_track])


def GetClientInstance(release_track=base.ReleaseTrack.GA, use_http=True):
  """Returns an instance of the Cloud Build client.

  Args:
    release_track: The desired value of the enum
      googlecloudsdk.calliope.base.ReleaseTrack.
    use_http: bool, True to create an http object for this client.

  Returns:
    base_api.BaseApiClient, An instance of the Cloud Build client.
  """
  return apis.GetClientInstance(
      _API_NAME,
      RELEASE_TRACK_TO_API_VERSION[release_track],
      no_http=(not use_http))


def EncodeSubstitutions(substitutions, messages):
  if not substitutions:
    return None
  # Sort for tests
  return apitools_encoding.DictToAdditionalPropertyMessage(
      substitutions, messages.Build.SubstitutionsValue, sort_items=True)


def EncodeTriggerSubstitutions(substitutions, messages):
  if not substitutions:
    return None
  substitution_properties = []
  for key, value in sorted(six.iteritems(substitutions)):  # Sort for tests
    substitution_properties.append(
        messages.BuildTrigger.SubstitutionsValue.AdditionalProperty(
            key=key, value=value))
  return messages.BuildTrigger.SubstitutionsValue(
      additionalProperties=substitution_properties)


def SnakeToCamelString(snake):
  """Change a snake_case string into a camelCase string.

  Args:
    snake: str, the string to be transformed.

  Returns:
    str, the transformed string.
  """
  parts = snake.split('_')
  if not parts:
    return snake

  # Handle snake with leading '_'s by collapsing them into the next part.
  # Legit field names will never look like this, but completeness of the
  # function is important.
  leading_blanks = 0
  for p in parts:
    if not p:
      leading_blanks += 1
    else:
      break
  if leading_blanks:
    parts = parts[leading_blanks:]
    if not parts:
      # If they were all blanks, then we over-counted by one because of split
      # behavior.
      return '_' * (leading_blanks - 1)
    parts[0] = '_' * leading_blanks + parts[0]

  return ''.join(parts[:1] + [s.capitalize() for s in parts[1:]])


def SnakeToCamel(msg, skip=None):
  """Recursively transform all keys and values from snake_case to camelCase.

  If a key is in skip, then its value is left alone.

  Args:
    msg: dict, list, or other. If 'other', the function returns immediately.
    skip: contains dict keys whose values should not have camel case applied.

  Returns:
    Same type as msg, except all strings that were snake_case are now CamelCase,
    except for the values of dict keys contained in skip.
  """
  if skip is None:
    skip = []
  if isinstance(msg, dict):
    return {
        SnakeToCamelString(key):
        (SnakeToCamel(val, skip) if key not in skip else val)
        for key, val in six.iteritems(msg)
    }
  elif isinstance(msg, list):
    return [SnakeToCamel(elem, skip) for elem in msg]
  else:
    return msg


def MessageToFieldPaths(msg):
  """Produce field paths from a message object.

  The result is used to create a FieldMask proto message that contains all field
  paths presented in the object.
  https://github.com/protocolbuffers/protobuf/blob/master/src/google/protobuf/field_mask.proto

  Args:
    msg: A user defined message object that extends the messages.Message class.
    https://github.com/google/apitools/blob/master/apitools/base/protorpclite/messages.py

  Returns:
    The list of field paths.
  """
  fields = []
  for field in msg.all_fields():
    v = msg.get_assigned_value(field.name)
    if field.repeated and not v:
      # Repeated field is initialized as an empty list.
      continue
    if v is not None:
      # ConvertToSnakeCase produces private_poolv1_config or hybrid_pool_config.
      if field.name == 'privatePoolV1Config':
        name = 'private_pool_v1_config'
      elif field.name == 'hybridPoolConfig':
        name = 'hybrid_pool_config'
      else:
        name = resource_property.ConvertToSnakeCase(field.name)
      if hasattr(v, 'all_fields'):
        # message has sub-messages, constructing subpaths.
        for f in MessageToFieldPaths(v):
          fields.append('{}.{}'.format(name, f))
      else:
        fields.append(name)
  return fields


def _UnpackCheckUnused(obj, msg_type):
  """Stuff a dict into a proto message, and fail if there are unused values.

  Args:
    obj: dict(), The structured data to be reflected into the message type.
    msg_type: type, The proto message type.

  Raises:
    ValueError: If there is an unused value in obj.

  Returns:
    Proto message, The message that was created from obj.
  """
  msg = apitools_encoding.DictToMessage(obj, msg_type)

  def _CheckForUnusedFields(obj):
    """Check for any unused fields in nested messages or lists."""
    if isinstance(obj, proto_messages.Message):
      unused_fields = obj.all_unrecognized_fields()
      if unused_fields:
        if len(unused_fields) > 1:
          # Because this message shows up in a dotted path, use braces.
          # eg .foo.bar.{x,y,z}
          unused_msg = '{%s}' % ','.join(sorted(unused_fields))
        else:
          # For single items, omit the braces.
          # eg .foo.bar.x
          unused_msg = unused_fields[0]
        raise ValueError('.%s: unused' % unused_msg)
      for used_field in obj.all_fields():
        try:
          field = getattr(obj, used_field.name)
          _CheckForUnusedFields(field)
        except ValueError as e:
          raise ValueError('.%s%s' % (used_field.name, e))
    if isinstance(obj, list):
      for i, item in enumerate(obj):
        try:
          _CheckForUnusedFields(item)
        except ValueError as e:
          raise ValueError('[%d]%s' % (i, e))

  _CheckForUnusedFields(msg)

  return msg


def LoadMessageFromStream(stream,
                          msg_type,
                          msg_friendly_name,
                          skip_camel_case=None,
                          path=None):
  """Load a proto message from a stream of JSON or YAML text.

  Args:
    stream: file-like object containing the JSON or YAML data to be decoded.
    msg_type: The protobuf message type to create.
    msg_friendly_name: A readable name for the message type, for use in error
      messages.
    skip_camel_case: Contains proto field names or map keys whose values should
      not have camel case applied.
    path: str or None. Optional path to be used in error messages.

  Raises:
    ParserError: If there was a problem parsing the stream as a dict.
    ParseProtoException: If there was a problem interpreting the stream as the
    given message type.

  Returns:
    Proto message, The message that got decoded.
  """
  if skip_camel_case is None:
    skip_camel_case = []
  # Turn the data into a dict
  try:
    structured_data = yaml.load(stream, file_hint=path)
  except yaml.Error as e:
    raise cloudbuild_exceptions.ParserError(path, e.inner_error)
  if not isinstance(structured_data, dict):
    raise cloudbuild_exceptions.ParserError(path,
                                            'Could not parse as a dictionary.')

  return _YamlToMessage(structured_data, msg_type, msg_friendly_name,
                        skip_camel_case, path)


def LoadMessagesFromStream(stream,
                           msg_type,
                           msg_friendly_name,
                           skip_camel_case=None,
                           path=None):
  """Load multiple proto message from a stream of JSON or YAML text.

  Args:
    stream: file-like object containing the JSON or YAML data to be decoded.
    msg_type: The protobuf message type to create.
    msg_friendly_name: A readable name for the message type, for use in error
      messages.
    skip_camel_case: Contains proto field names or map keys whose values should
      not have camel case applied.
    path: str or None. Optional path to be used in error messages.

  Raises:
    ParserError: If there was a problem parsing the stream.
    ParseProtoException: If there was a problem interpreting the stream as the
    given message type.

  Returns:
    Proto message list of the messages that got decoded.
  """
  if skip_camel_case is None:
    skip_camel_case = []
  # Turn the data into a dict
  try:
    structured_data = yaml.load_all(stream, file_hint=path)
  except yaml.Error as e:
    raise cloudbuild_exceptions.ParserError(path, e.inner_error)

  return [
      _YamlToMessage(item, msg_type, msg_friendly_name, skip_camel_case, path)
      for item in structured_data
  ]


def _YamlToMessage(structured_data,
                   msg_type,
                   msg_friendly_name,
                   skip_camel_case=None,
                   path=None):
  """Load a proto message from a file containing JSON or YAML text.

  Args:
    structured_data: Dict containing the decoded YAML data.
    msg_type: The protobuf message type to create.
    msg_friendly_name: A readable name for the message type, for use in error
      messages.
    skip_camel_case: Contains proto field names or map keys whose values should
      not have camel case applied.
    path: str or None. Optional path to be used in error messages.

  Raises:
    ParseProtoException: If there was a problem interpreting the file as the
    given message type.

  Returns:
    Proto message, The message that got decoded.
  """

  # Transform snake_case into camelCase.
  structured_data = SnakeToCamel(structured_data, skip_camel_case)

  # Then, turn the dict into a proto message.
  try:
    msg = _UnpackCheckUnused(structured_data, msg_type)
  except Exception as e:
    # Catch all exceptions here because a valid YAML can sometimes not be a
    # valid message, so we need to catch all errors in the dict to message
    # conversion.
    raise cloudbuild_exceptions.ParseProtoException(path, msg_friendly_name,
                                                    '%s' % e)

  return msg


def LoadMessageFromPath(path,
                        msg_type,
                        msg_friendly_name,
                        skip_camel_case=None):
  """Load a proto message from a file containing JSON or YAML text.

  Args:
    path: The path to a file containing the JSON or YAML data to be decoded.
    msg_type: The protobuf message type to create.
    msg_friendly_name: A readable name for the message type, for use in error
      messages.
    skip_camel_case: Contains proto field names or map keys whose values should
      not have camel case applied.

  Raises:
    files.MissingFileError: If the file does not exist.
    ParserError: If there was a problem parsing the file as a dict.
    ParseProtoException: If there was a problem interpreting the file as the
    given message type.

  Returns:
    Proto message, The message that got decoded.
  """
  with files.FileReader(path) as f:  # Returns user-friendly error messages
    return LoadMessageFromStream(f, msg_type, msg_friendly_name,
                                 skip_camel_case, path)


def LoadMessagesFromPath(path,
                         msg_type,
                         msg_friendly_name,
                         skip_camel_case=None):
  """Load a proto message from a file containing JSON or YAML text.

  Args:
    path: The path to a file containing the JSON or YAML data to be decoded.
    msg_type: The protobuf message type to create.
    msg_friendly_name: A readable name for the message type, for use in error
      messages.
    skip_camel_case: Contains proto field names or map keys whose values should
      not have camel case applied.

  Raises:
    files.MissingFileError: If the file does not exist.
    ParseProtoException: If there was a problem interpreting the file as the
    given message type.

  Returns:
    Proto message list of the messages that got decoded.
  """
  with files.FileReader(path) as f:  # Returns user-friendly error messages
    return LoadMessagesFromStream(f, msg_type, msg_friendly_name,
                                  skip_camel_case, path)


def IsWorkerPool(resource_name):
  """Determine if the provided full resource name is a worker pool.

  Args:
    resource_name: str, The string to test.

  Returns:
    bool, True if the string is a worker pool's full resource name.
  """
  return bool(re.match(WORKERPOOL_NAME_MATCHER, resource_name))


def WorkerPoolShortName(resource_name):
  """Get the name part of a worker pool's full resource name.

  For example, "projects/abc/locations/def/workerPools/ghi" returns "ghi".

  Args:
    resource_name: A worker pool's full resource name.

  Raises:
    ValueError: If the full resource name was not well-formatted.

  Returns:
    The worker pool's short name.
  """
  match = re.search(WORKERPOOL_NAME_SELECTOR, resource_name)
  if match:
    return match.group(1)
  raise ValueError('The worker pool resource name must match "%s"' %
                   (WORKERPOOL_NAME_MATCHER,))


def WorkerPoolRegion(resource_name):
  """Get the region part of a worker pool's full resource name.

  For example, "projects/abc/locations/def/workerPools/ghi" returns "def".

  Args:
    resource_name: str, A worker pool's full resource name.

  Raises:
    ValueError: If the full resource name was not well-formatted.

  Returns:
    str, The worker pool's region string.
  """
  match = re.search(WORKERPOOL_REGION_SELECTOR, resource_name)
  if match:
    return match.group(1)
  raise ValueError('The worker pool resource name must match "%s"' %
                   (WORKERPOOL_NAME_MATCHER,))


def GitHubEnterpriseConfigFromArgs(args, update=False):
  """Construct the GitHubEnterpires resource from the command line args.

  Args:
    args: an argparse namespace. All the arguments that were provided to this
        command invocation.
      update: bool, if the args are for an update.

  Returns:
    A populated GitHubEnterpriseConfig message.
  """
  messages = GetMessagesModule()

  ghe = messages.GitHubEnterpriseConfig()
  ghe.hostUrl = args.host_uri
  ghe.appId = args.app_id
  if args.webhook_key is not None:
    ghe.webhookKey = args.webhook_key
  if not update and args.peered_network is not None:
    ghe.peeredNetwork = args.peered_network
  if args.gcs_bucket is not None:
    gcs_location = messages.GCSLocation()
    gcs_location.bucket = args.gcs_bucket
    gcs_location.object = args.gcs_object
    if args.generation is not None:
      gcs_location.generation = args.generation
    ghe.appConfigJson = gcs_location
  else:
    secret_location = messages.GitHubEnterpriseSecrets()
    secret_location.privateKeyName = args.private_key_name
    secret_location.webhookSecretName = args.webhook_secret_name
    secret_location.oauthSecretName = args.oauth_secret_name
    secret_location.oauthClientIdName = args.oauth_client_id_name
    ghe.secrets = secret_location
  return ghe


def BitbucketServerConfigFromArgs(args, update=False):
  """Construct the BitbucketServer resource from the command line args.

  Args:
    args: an argparse namespace. All the arguments that were provided to this
      command invocation.
    update: bool, if the args are for an update.

  Returns:
    A populated BitbucketServerConfig message.
  """
  messages = GetMessagesModule()

  bbs = messages.BitbucketServerConfig()
  bbs.hostUri = args.host_uri
  bbs.username = args.user_name
  bbs.apiKey = args.api_key
  secret_location = messages.BitbucketServerSecrets()
  secret_location.adminAccessTokenVersionName = args.admin_access_token_secret_version
  secret_location.readAccessTokenVersionName = args.read_access_token_secret_version
  secret_location.webhookSecretVersionName = args.webhook_secret_secret_version
  if update or secret_location is not None:
    bbs.secrets = secret_location
  if not update and args.peered_network is not None:
    bbs.peeredNetwork = args.peered_network
  if args.IsSpecified('ssl_ca_file'):
    bbs.sslCa = args.ssl_ca_file
  return bbs


def GitLabConfigFromArgs(args):
  """Construct the GitLabConfig resource from the command line args.

  Args:
    args: an argparse namespace. All the arguments that were provided to this
      command invocation.

  Returns:
    A populated GitLabConfig message.
  """
  messages = GetMessagesModule()

  config = messages.GitLabConfig()
  config.username = args.user_name

  secrets = messages.GitLabSecrets()
  secrets.apiAccessTokenVersion = args.api_access_token_secret_version
  secrets.readAccessTokenVersion = args.read_access_token_secret_version
  secrets.webhookSecretVersion = args.webhook_secret_secret_version
  secrets.apiKeyVersion = args.api_key_secret_version
  if not _IsEmptyMessage(secrets):
    config.secrets = secrets

  enterprise_config = messages.GitLabEnterpriseConfig()
  enterprise_config.hostUri = args.host_uri
  service_directory_config = messages.ServiceDirectoryConfig()
  service_directory_config.service = args.service_directory_service
  enterprise_config.serviceDirectoryConfig = service_directory_config
  if args.IsSpecified('ssl_ca_file'):
    enterprise_config.sslCa = args.ssl_ca_file
  if not _IsEmptyMessage(enterprise_config):
    config.enterpriseConfig = enterprise_config

  return config


def _IsEmptyMessage(message):
  if message is None:
    return True

  message_dict = apitools_encoding.MessageToDict(message)
  return not any(message_dict.values())


def WorkerPoolIsSpecified(build_config):
  return build_config is not None and build_config.options is not None and build_config.options.pool is not None and build_config.options.pool.name is not None


def WorkerPoolConfigIsSpecified(build_config):
  return build_config is not None and build_config.options is not None and build_config.options.pool is not None and build_config.options.pool.workerConfig is not None


def BytesToGb(size):
  """Converts bytes to GB.

  Args:
    size: a size in GB  Does not require size to be a multiple of 1 GB unlike
      utils.BytesToGb from from googlecloudsdk.api_lib.compute

  Returns:
    size in bytes.
  """
  if not size:
    return None

  return size // BYTES_IN_ONE_GB



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Base SSHTroubleshooter class for each category of ssh troubleshooting check."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

import abc

import six


@six.add_metaclass(abc.ABCMeta)
class SshTroubleshooter(object):
  """A class whose instance is a ssh troubleshooter.

  Test authors should subclass Troubleshooter for each troubleshooter
  subcategory.
  """

  project = None
  zone = None
  instance = None

  def __init__(self):
    """Initialize with project and instance object, and zone.
    """
    raise NotImplementedError

  @abc.abstractmethod
  def check_prerequisite(self):
    """Hook method for checking prerequisite for troubleshooting before troubleshoot action.
    """
    raise NotImplementedError

  @abc.abstractmethod
  def cleanup_resources(self):
    """Hook method for cleaning troubleshooting resource after troubleshooting action.
    """
    raise NotImplementedError

  @abc.abstractmethod
  def troubleshoot(self):
    """Hook method for troubleshooting action."""
    raise NotImplementedError

  def __call__(self):
    self.run()

  def run(self):
    self.check_prerequisite()
    self.troubleshoot()
    self.cleanup_resources()



# -*- coding: utf-8 -*- #
# Copyright 2013 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Revoke credentials being used by the CloudSDK."""

from __future__ import absolute_import
from __future__ import division
from __future__ import unicode_literals

from googlecloudsdk.calliope import base
from googlecloudsdk.calliope import exceptions as c_exc
from googlecloudsdk.command_lib.auth import auth_util
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core.credentials import creds as c_creds
from googlecloudsdk.core.credentials import exceptions as creds_exceptions
from googlecloudsdk.core.credentials import store as c_store
from googlecloudsdk.core.resource import resource_printer


class Revoke(base.Command):
  """Revoke access credentials for an account.

  Revokes credentials for the specified user accounts, service accounts or
  external accounts (workload identity pools).

  When given a user account, this command revokes the user account token on the
  server. If the revocation is successful, or if the token has already been
  revoked, this command removes the credential from the local machine.

  When given a service account, this command does not revoke the service account
  token on the server because service account tokens are not revocable. Instead,
  it will print a warning and remove the credential from the local machine. When
  used with a service account, this command has only a local effect and the key
  associated with the service account is not deleted. This can be done by
  executing `gcloud iam service-accounts keys delete` after `revoke`.

  When given an external account (workload identity pool), whether impersonated
  or not, the command does not revoke the corresponding token on the server
  because these tokens are not revocable. The underlying external credentials
  (OIDC, AWS, etc.) used to generate these access tokens have to be revoked too,
  but gcloud has no control over that. Instead, it will print a warning and
  remove the credential from the local machine.

  If no account is specified, this command revokes credentials for the currently
  active account, effectively logging out of that account. If --all is given,
  the behaviors described above apply individually to each account in the list.

  You can revoke credentials when you want to prevent gcloud and other Google
  Cloud CLI tools from using the specified account. You do not need to revoke
  credentials to switch between accounts.
  """

  @staticmethod
  def Args(parser):
    parser.add_argument('accounts', nargs='*',
                        help='Accounts whose credentials are to be revoked.')
    parser.add_argument('--all', action='store_true',
                        help='Revoke credentials for all accounts.')
    parser.display_info.AddFormat('list[title="Revoked credentials:"]')

  def Run(self, args):
    """Revoke credentials and update active account."""
    accounts = args.accounts or []
    if isinstance(accounts, str):
      accounts = [accounts]
    available_accounts = c_store.AvailableAccounts()
    unknown_accounts = set(accounts) - set(available_accounts)
    if unknown_accounts:
      raise c_exc.UnknownArgumentException(
          'accounts', ' '.join(unknown_accounts))
    if args.all:
      accounts = available_accounts

    active_account = properties.VALUES.core.account.Get()

    if not accounts and active_account:
      accounts = [active_account]

    if not accounts:
      raise c_exc.InvalidArgumentException(
          'accounts', 'No credentials available to revoke.')

    for account in accounts:
      if active_account == account:
        properties.PersistProperty(properties.VALUES.core.account, None)
      # External account and external account user credentials cannot be
      # revoked.
      # Detect these type of credentials to show a more user friendly message
      # on revocation calls.
      # Note that impersonated external account credentials will appear like
      # service accounts. These will end with gserviceaccount.com and will be
      # handled the same way service account credentials are handled.
      try:
        creds = c_store.Load(
            account, prevent_refresh=True, use_google_auth=True)
      except creds_exceptions.Error:
        # Ignore all errors. These will be properly handled in the subsequent
        # Revoke call.
        creds = None
      if not c_store.Revoke(account):
        if account.endswith('.gserviceaccount.com'):
          log.warning(
              '[{}] appears to be a service account. Service account tokens '
              'cannot be revoked, but they will expire automatically. To '
              'prevent use of the service account token earlier than the '
              'expiration, delete or disable the parent service account. To '
              'explicitly delete the key associated with the service account '
              'use `gcloud iam service-accounts keys delete` instead`.'.format(
                  account))
        elif c_creds.IsExternalAccountCredentials(creds):
          log.warning(
              '[{}] appears to be an external account. External account '
              'tokens cannot be revoked, but they will expire automatically.'
              .format(account))
        elif c_creds.IsExternalAccountUserCredentials(creds):
          log.warning(
              '[{}] appears to be an external account user. External account '
              'user tokens cannot be revoked, but they will expire '
              'automatically.'.format(account))
        else:
          log.warning(
              '[{}] already inactive (previously revoked?)'.format(account))
    return accounts

  def Epilog(self, unused_results_were_displayed):
    accounts = auth_util.AllAccounts()
    printer = resource_printer.Printer(
        auth_util.ACCOUNT_TABLE_FORMAT,
        out=log.status)
    printer.Print(accounts)



"""Generated client library for apigateway version v1beta."""
# NOTE: This file is autogenerated and should not be edited by hand.

from __future__ import absolute_import

from apitools.base.py import base_api
from googlecloudsdk.third_party.apis.apigateway.v1beta import apigateway_v1beta_messages as messages


class ApigatewayV1beta(base_api.BaseApiClient):
  """Generated client library for service apigateway version v1beta."""

  MESSAGES_MODULE = messages
  BASE_URL = 'https://apigateway.googleapis.com/'
  MTLS_BASE_URL = 'https://apigateway.mtls.googleapis.com/'

  _PACKAGE = 'apigateway'
  _SCOPES = ['https://www.googleapis.com/auth/cloud-platform']
  _VERSION = 'v1beta'
  _CLIENT_ID = 'CLIENT_ID'
  _CLIENT_SECRET = 'CLIENT_SECRET'
  _USER_AGENT = 'google-cloud-sdk'
  _CLIENT_CLASS_NAME = 'ApigatewayV1beta'
  _URL_VERSION = 'v1beta'
  _API_KEY = None

  def __init__(self, url='', credentials=None,
               get_credentials=True, http=None, model=None,
               log_request=False, log_response=False,
               credentials_args=None, default_global_params=None,
               additional_http_headers=None, response_encoding=None):
    """Create a new apigateway handle."""
    url = url or self.BASE_URL
    super(ApigatewayV1beta, self).__init__(
        url, credentials=credentials,
        get_credentials=get_credentials, http=http, model=model,
        log_request=log_request, log_response=log_response,
        credentials_args=credentials_args,
        default_global_params=default_global_params,
        additional_http_headers=additional_http_headers,
        response_encoding=response_encoding)
    self.projects_locations_apis_configs = self.ProjectsLocationsApisConfigsService(self)
    self.projects_locations_apis = self.ProjectsLocationsApisService(self)
    self.projects_locations_gateways = self.ProjectsLocationsGatewaysService(self)
    self.projects_locations_operations = self.ProjectsLocationsOperationsService(self)
    self.projects_locations = self.ProjectsLocationsService(self)
    self.projects = self.ProjectsService(self)

  class ProjectsLocationsApisConfigsService(base_api.BaseApiService):
    """Service class for the projects_locations_apis_configs resource."""

    _NAME = 'projects_locations_apis_configs'

    def __init__(self, client):
      super(ApigatewayV1beta.ProjectsLocationsApisConfigsService, self).__init__(client)
      self._upload_configs = {
          }

    def Create(self, request, global_params=None):
      r"""Creates a new ApiConfig in a given project and location.

      Args:
        request: (ApigatewayProjectsLocationsApisConfigsCreateRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayOperation) The response message.
      """
      config = self.GetMethodConfig('Create')
      return self._RunMethod(
          config, request, global_params=global_params)

    Create.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}/configs',
        http_method='POST',
        method_id='apigateway.projects.locations.apis.configs.create',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['apiConfigId'],
        relative_path='v1beta/{+parent}/configs',
        request_field='apigatewayApiConfig',
        request_type_name='ApigatewayProjectsLocationsApisConfigsCreateRequest',
        response_type_name='ApigatewayOperation',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes a single ApiConfig.

      Args:
        request: (ApigatewayProjectsLocationsApisConfigsDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayOperation) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}/configs/{configsId}',
        http_method='DELETE',
        method_id='apigateway.projects.locations.apis.configs.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1beta/{+name}',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsApisConfigsDeleteRequest',
        response_type_name='ApigatewayOperation',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets details of a single ApiConfig.

      Args:
        request: (ApigatewayProjectsLocationsApisConfigsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayApiConfig) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}/configs/{configsId}',
        http_method='GET',
        method_id='apigateway.projects.locations.apis.configs.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['view'],
        relative_path='v1beta/{+name}',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsApisConfigsGetRequest',
        response_type_name='ApigatewayApiConfig',
        supports_download=False,
    )

    def GetIamPolicy(self, request, global_params=None):
      r"""Gets the access control policy for a resource. Returns an empty policy if the resource exists and does not have a policy set.

      Args:
        request: (ApigatewayProjectsLocationsApisConfigsGetIamPolicyRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayPolicy) The response message.
      """
      config = self.GetMethodConfig('GetIamPolicy')
      return self._RunMethod(
          config, request, global_params=global_params)

    GetIamPolicy.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}/configs/{configsId}:getIamPolicy',
        http_method='GET',
        method_id='apigateway.projects.locations.apis.configs.getIamPolicy',
        ordered_params=['resource'],
        path_params=['resource'],
        query_params=['options_requestedPolicyVersion'],
        relative_path='v1beta/{+resource}:getIamPolicy',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsApisConfigsGetIamPolicyRequest',
        response_type_name='ApigatewayPolicy',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists ApiConfigs in a given project and location.

      Args:
        request: (ApigatewayProjectsLocationsApisConfigsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayListApiConfigsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}/configs',
        http_method='GET',
        method_id='apigateway.projects.locations.apis.configs.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['filter', 'orderBy', 'pageSize', 'pageToken'],
        relative_path='v1beta/{+parent}/configs',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsApisConfigsListRequest',
        response_type_name='ApigatewayListApiConfigsResponse',
        supports_download=False,
    )

    def Patch(self, request, global_params=None):
      r"""Updates the parameters of a single ApiConfig.

      Args:
        request: (ApigatewayProjectsLocationsApisConfigsPatchRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayOperation) The response message.
      """
      config = self.GetMethodConfig('Patch')
      return self._RunMethod(
          config, request, global_params=global_params)

    Patch.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}/configs/{configsId}',
        http_method='PATCH',
        method_id='apigateway.projects.locations.apis.configs.patch',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['updateMask'],
        relative_path='v1beta/{+name}',
        request_field='apigatewayApiConfig',
        request_type_name='ApigatewayProjectsLocationsApisConfigsPatchRequest',
        response_type_name='ApigatewayOperation',
        supports_download=False,
    )

    def SetIamPolicy(self, request, global_params=None):
      r"""Sets the access control policy on the specified resource. Replaces any existing policy. Can return `NOT_FOUND`, `INVALID_ARGUMENT`, and `PERMISSION_DENIED` errors.

      Args:
        request: (ApigatewayProjectsLocationsApisConfigsSetIamPolicyRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayPolicy) The response message.
      """
      config = self.GetMethodConfig('SetIamPolicy')
      return self._RunMethod(
          config, request, global_params=global_params)

    SetIamPolicy.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}/configs/{configsId}:setIamPolicy',
        http_method='POST',
        method_id='apigateway.projects.locations.apis.configs.setIamPolicy',
        ordered_params=['resource'],
        path_params=['resource'],
        query_params=[],
        relative_path='v1beta/{+resource}:setIamPolicy',
        request_field='apigatewaySetIamPolicyRequest',
        request_type_name='ApigatewayProjectsLocationsApisConfigsSetIamPolicyRequest',
        response_type_name='ApigatewayPolicy',
        supports_download=False,
    )

    def TestIamPermissions(self, request, global_params=None):
      r"""Returns permissions that a caller has on the specified resource. If the resource does not exist, this will return an empty set of permissions, not a `NOT_FOUND` error. Note: This operation is designed to be used for building permission-aware UIs and command-line tools, not for authorization checking. This operation may "fail open" without warning.

      Args:
        request: (ApigatewayProjectsLocationsApisConfigsTestIamPermissionsRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayTestIamPermissionsResponse) The response message.
      """
      config = self.GetMethodConfig('TestIamPermissions')
      return self._RunMethod(
          config, request, global_params=global_params)

    TestIamPermissions.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}/configs/{configsId}:testIamPermissions',
        http_method='POST',
        method_id='apigateway.projects.locations.apis.configs.testIamPermissions',
        ordered_params=['resource'],
        path_params=['resource'],
        query_params=[],
        relative_path='v1beta/{+resource}:testIamPermissions',
        request_field='apigatewayTestIamPermissionsRequest',
        request_type_name='ApigatewayProjectsLocationsApisConfigsTestIamPermissionsRequest',
        response_type_name='ApigatewayTestIamPermissionsResponse',
        supports_download=False,
    )

  class ProjectsLocationsApisService(base_api.BaseApiService):
    """Service class for the projects_locations_apis resource."""

    _NAME = 'projects_locations_apis'

    def __init__(self, client):
      super(ApigatewayV1beta.ProjectsLocationsApisService, self).__init__(client)
      self._upload_configs = {
          }

    def Create(self, request, global_params=None):
      r"""Creates a new Api in a given project and location.

      Args:
        request: (ApigatewayProjectsLocationsApisCreateRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayOperation) The response message.
      """
      config = self.GetMethodConfig('Create')
      return self._RunMethod(
          config, request, global_params=global_params)

    Create.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis',
        http_method='POST',
        method_id='apigateway.projects.locations.apis.create',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['apiId'],
        relative_path='v1beta/{+parent}/apis',
        request_field='apigatewayApi',
        request_type_name='ApigatewayProjectsLocationsApisCreateRequest',
        response_type_name='ApigatewayOperation',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes a single Api.

      Args:
        request: (ApigatewayProjectsLocationsApisDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayOperation) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}',
        http_method='DELETE',
        method_id='apigateway.projects.locations.apis.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1beta/{+name}',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsApisDeleteRequest',
        response_type_name='ApigatewayOperation',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets details of a single Api.

      Args:
        request: (ApigatewayProjectsLocationsApisGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayApi) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}',
        http_method='GET',
        method_id='apigateway.projects.locations.apis.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1beta/{+name}',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsApisGetRequest',
        response_type_name='ApigatewayApi',
        supports_download=False,
    )

    def GetIamPolicy(self, request, global_params=None):
      r"""Gets the access control policy for a resource. Returns an empty policy if the resource exists and does not have a policy set.

      Args:
        request: (ApigatewayProjectsLocationsApisGetIamPolicyRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayPolicy) The response message.
      """
      config = self.GetMethodConfig('GetIamPolicy')
      return self._RunMethod(
          config, request, global_params=global_params)

    GetIamPolicy.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}:getIamPolicy',
        http_method='GET',
        method_id='apigateway.projects.locations.apis.getIamPolicy',
        ordered_params=['resource'],
        path_params=['resource'],
        query_params=['options_requestedPolicyVersion'],
        relative_path='v1beta/{+resource}:getIamPolicy',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsApisGetIamPolicyRequest',
        response_type_name='ApigatewayPolicy',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists Apis in a given project and location.

      Args:
        request: (ApigatewayProjectsLocationsApisListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayListApisResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis',
        http_method='GET',
        method_id='apigateway.projects.locations.apis.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['filter', 'orderBy', 'pageSize', 'pageToken'],
        relative_path='v1beta/{+parent}/apis',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsApisListRequest',
        response_type_name='ApigatewayListApisResponse',
        supports_download=False,
    )

    def Patch(self, request, global_params=None):
      r"""Updates the parameters of a single Api.

      Args:
        request: (ApigatewayProjectsLocationsApisPatchRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayOperation) The response message.
      """
      config = self.GetMethodConfig('Patch')
      return self._RunMethod(
          config, request, global_params=global_params)

    Patch.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}',
        http_method='PATCH',
        method_id='apigateway.projects.locations.apis.patch',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['updateMask'],
        relative_path='v1beta/{+name}',
        request_field='apigatewayApi',
        request_type_name='ApigatewayProjectsLocationsApisPatchRequest',
        response_type_name='ApigatewayOperation',
        supports_download=False,
    )

    def SetIamPolicy(self, request, global_params=None):
      r"""Sets the access control policy on the specified resource. Replaces any existing policy. Can return `NOT_FOUND`, `INVALID_ARGUMENT`, and `PERMISSION_DENIED` errors.

      Args:
        request: (ApigatewayProjectsLocationsApisSetIamPolicyRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayPolicy) The response message.
      """
      config = self.GetMethodConfig('SetIamPolicy')
      return self._RunMethod(
          config, request, global_params=global_params)

    SetIamPolicy.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}:setIamPolicy',
        http_method='POST',
        method_id='apigateway.projects.locations.apis.setIamPolicy',
        ordered_params=['resource'],
        path_params=['resource'],
        query_params=[],
        relative_path='v1beta/{+resource}:setIamPolicy',
        request_field='apigatewaySetIamPolicyRequest',
        request_type_name='ApigatewayProjectsLocationsApisSetIamPolicyRequest',
        response_type_name='ApigatewayPolicy',
        supports_download=False,
    )

    def TestIamPermissions(self, request, global_params=None):
      r"""Returns permissions that a caller has on the specified resource. If the resource does not exist, this will return an empty set of permissions, not a `NOT_FOUND` error. Note: This operation is designed to be used for building permission-aware UIs and command-line tools, not for authorization checking. This operation may "fail open" without warning.

      Args:
        request: (ApigatewayProjectsLocationsApisTestIamPermissionsRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayTestIamPermissionsResponse) The response message.
      """
      config = self.GetMethodConfig('TestIamPermissions')
      return self._RunMethod(
          config, request, global_params=global_params)

    TestIamPermissions.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/apis/{apisId}:testIamPermissions',
        http_method='POST',
        method_id='apigateway.projects.locations.apis.testIamPermissions',
        ordered_params=['resource'],
        path_params=['resource'],
        query_params=[],
        relative_path='v1beta/{+resource}:testIamPermissions',
        request_field='apigatewayTestIamPermissionsRequest',
        request_type_name='ApigatewayProjectsLocationsApisTestIamPermissionsRequest',
        response_type_name='ApigatewayTestIamPermissionsResponse',
        supports_download=False,
    )

  class ProjectsLocationsGatewaysService(base_api.BaseApiService):
    """Service class for the projects_locations_gateways resource."""

    _NAME = 'projects_locations_gateways'

    def __init__(self, client):
      super(ApigatewayV1beta.ProjectsLocationsGatewaysService, self).__init__(client)
      self._upload_configs = {
          }

    def Create(self, request, global_params=None):
      r"""Creates a new Gateway in a given project and location.

      Args:
        request: (ApigatewayProjectsLocationsGatewaysCreateRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayOperation) The response message.
      """
      config = self.GetMethodConfig('Create')
      return self._RunMethod(
          config, request, global_params=global_params)

    Create.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/gateways',
        http_method='POST',
        method_id='apigateway.projects.locations.gateways.create',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['gatewayId'],
        relative_path='v1beta/{+parent}/gateways',
        request_field='apigatewayGateway',
        request_type_name='ApigatewayProjectsLocationsGatewaysCreateRequest',
        response_type_name='ApigatewayOperation',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes a single Gateway.

      Args:
        request: (ApigatewayProjectsLocationsGatewaysDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayOperation) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/gateways/{gatewaysId}',
        http_method='DELETE',
        method_id='apigateway.projects.locations.gateways.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1beta/{+name}',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsGatewaysDeleteRequest',
        response_type_name='ApigatewayOperation',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets details of a single Gateway.

      Args:
        request: (ApigatewayProjectsLocationsGatewaysGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayGateway) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/gateways/{gatewaysId}',
        http_method='GET',
        method_id='apigateway.projects.locations.gateways.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1beta/{+name}',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsGatewaysGetRequest',
        response_type_name='ApigatewayGateway',
        supports_download=False,
    )

    def GetIamPolicy(self, request, global_params=None):
      r"""Gets the access control policy for a resource. Returns an empty policy if the resource exists and does not have a policy set.

      Args:
        request: (ApigatewayProjectsLocationsGatewaysGetIamPolicyRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayPolicy) The response message.
      """
      config = self.GetMethodConfig('GetIamPolicy')
      return self._RunMethod(
          config, request, global_params=global_params)

    GetIamPolicy.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/gateways/{gatewaysId}:getIamPolicy',
        http_method='GET',
        method_id='apigateway.projects.locations.gateways.getIamPolicy',
        ordered_params=['resource'],
        path_params=['resource'],
        query_params=['options_requestedPolicyVersion'],
        relative_path='v1beta/{+resource}:getIamPolicy',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsGatewaysGetIamPolicyRequest',
        response_type_name='ApigatewayPolicy',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists Gateways in a given project and location.

      Args:
        request: (ApigatewayProjectsLocationsGatewaysListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayListGatewaysResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/gateways',
        http_method='GET',
        method_id='apigateway.projects.locations.gateways.list',
        ordered_params=['parent'],
        path_params=['parent'],
        query_params=['filter', 'orderBy', 'pageSize', 'pageToken'],
        relative_path='v1beta/{+parent}/gateways',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsGatewaysListRequest',
        response_type_name='ApigatewayListGatewaysResponse',
        supports_download=False,
    )

    def Patch(self, request, global_params=None):
      r"""Updates the parameters of a single Gateway.

      Args:
        request: (ApigatewayProjectsLocationsGatewaysPatchRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayOperation) The response message.
      """
      config = self.GetMethodConfig('Patch')
      return self._RunMethod(
          config, request, global_params=global_params)

    Patch.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/gateways/{gatewaysId}',
        http_method='PATCH',
        method_id='apigateway.projects.locations.gateways.patch',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['updateMask'],
        relative_path='v1beta/{+name}',
        request_field='apigatewayGateway',
        request_type_name='ApigatewayProjectsLocationsGatewaysPatchRequest',
        response_type_name='ApigatewayOperation',
        supports_download=False,
    )

    def SetIamPolicy(self, request, global_params=None):
      r"""Sets the access control policy on the specified resource. Replaces any existing policy. Can return `NOT_FOUND`, `INVALID_ARGUMENT`, and `PERMISSION_DENIED` errors.

      Args:
        request: (ApigatewayProjectsLocationsGatewaysSetIamPolicyRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayPolicy) The response message.
      """
      config = self.GetMethodConfig('SetIamPolicy')
      return self._RunMethod(
          config, request, global_params=global_params)

    SetIamPolicy.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/gateways/{gatewaysId}:setIamPolicy',
        http_method='POST',
        method_id='apigateway.projects.locations.gateways.setIamPolicy',
        ordered_params=['resource'],
        path_params=['resource'],
        query_params=[],
        relative_path='v1beta/{+resource}:setIamPolicy',
        request_field='apigatewaySetIamPolicyRequest',
        request_type_name='ApigatewayProjectsLocationsGatewaysSetIamPolicyRequest',
        response_type_name='ApigatewayPolicy',
        supports_download=False,
    )

    def TestIamPermissions(self, request, global_params=None):
      r"""Returns permissions that a caller has on the specified resource. If the resource does not exist, this will return an empty set of permissions, not a `NOT_FOUND` error. Note: This operation is designed to be used for building permission-aware UIs and command-line tools, not for authorization checking. This operation may "fail open" without warning.

      Args:
        request: (ApigatewayProjectsLocationsGatewaysTestIamPermissionsRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayTestIamPermissionsResponse) The response message.
      """
      config = self.GetMethodConfig('TestIamPermissions')
      return self._RunMethod(
          config, request, global_params=global_params)

    TestIamPermissions.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/gateways/{gatewaysId}:testIamPermissions',
        http_method='POST',
        method_id='apigateway.projects.locations.gateways.testIamPermissions',
        ordered_params=['resource'],
        path_params=['resource'],
        query_params=[],
        relative_path='v1beta/{+resource}:testIamPermissions',
        request_field='apigatewayTestIamPermissionsRequest',
        request_type_name='ApigatewayProjectsLocationsGatewaysTestIamPermissionsRequest',
        response_type_name='ApigatewayTestIamPermissionsResponse',
        supports_download=False,
    )

  class ProjectsLocationsOperationsService(base_api.BaseApiService):
    """Service class for the projects_locations_operations resource."""

    _NAME = 'projects_locations_operations'

    def __init__(self, client):
      super(ApigatewayV1beta.ProjectsLocationsOperationsService, self).__init__(client)
      self._upload_configs = {
          }

    def Cancel(self, request, global_params=None):
      r"""Starts asynchronous cancellation on a long-running operation. The server makes a best effort to cancel the operation, but success is not guaranteed. If the server doesn't support this method, it returns `google.rpc.Code.UNIMPLEMENTED`. Clients can use Operations.GetOperation or other methods to check whether the cancellation succeeded or whether the operation completed despite cancellation. On successful cancellation, the operation is not deleted; instead, it becomes an operation with an Operation.error value with a google.rpc.Status.code of 1, corresponding to `Code.CANCELLED`.

      Args:
        request: (ApigatewayProjectsLocationsOperationsCancelRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Cancel')
      return self._RunMethod(
          config, request, global_params=global_params)

    Cancel.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/operations/{operationsId}:cancel',
        http_method='POST',
        method_id='apigateway.projects.locations.operations.cancel',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1beta/{+name}:cancel',
        request_field='apigatewayCancelOperationRequest',
        request_type_name='ApigatewayProjectsLocationsOperationsCancelRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Delete(self, request, global_params=None):
      r"""Deletes a long-running operation. This method indicates that the client is no longer interested in the operation result. It does not cancel the operation. If the server doesn't support this method, it returns `google.rpc.Code.UNIMPLEMENTED`.

      Args:
        request: (ApigatewayProjectsLocationsOperationsDeleteRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (Empty) The response message.
      """
      config = self.GetMethodConfig('Delete')
      return self._RunMethod(
          config, request, global_params=global_params)

    Delete.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/operations/{operationsId}',
        http_method='DELETE',
        method_id='apigateway.projects.locations.operations.delete',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1beta/{+name}',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsOperationsDeleteRequest',
        response_type_name='Empty',
        supports_download=False,
    )

    def Get(self, request, global_params=None):
      r"""Gets the latest state of a long-running operation. Clients can use this method to poll the operation result at intervals as recommended by the API service.

      Args:
        request: (ApigatewayProjectsLocationsOperationsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayOperation) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/operations/{operationsId}',
        http_method='GET',
        method_id='apigateway.projects.locations.operations.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1beta/{+name}',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsOperationsGetRequest',
        response_type_name='ApigatewayOperation',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists operations that match the specified filter in the request. If the server doesn't support this method, it returns `UNIMPLEMENTED`. NOTE: the `name` binding allows API services to override the binding to use different resource name schemes, such as `users/*/operations`. To override the binding, API services can add a binding such as `"/v1/{name=users/*}/operations"` to their service configuration. For backwards compatibility, the default name includes the operations collection id, however overriding users must ensure the name binding is the parent resource, without the operations collection id.

      Args:
        request: (ApigatewayProjectsLocationsOperationsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayListOperationsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}/operations',
        http_method='GET',
        method_id='apigateway.projects.locations.operations.list',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['filter', 'pageSize', 'pageToken'],
        relative_path='v1beta/{+name}/operations',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsOperationsListRequest',
        response_type_name='ApigatewayListOperationsResponse',
        supports_download=False,
    )

  class ProjectsLocationsService(base_api.BaseApiService):
    """Service class for the projects_locations resource."""

    _NAME = 'projects_locations'

    def __init__(self, client):
      super(ApigatewayV1beta.ProjectsLocationsService, self).__init__(client)
      self._upload_configs = {
          }

    def Get(self, request, global_params=None):
      r"""Gets information about a location.

      Args:
        request: (ApigatewayProjectsLocationsGetRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayLocation) The response message.
      """
      config = self.GetMethodConfig('Get')
      return self._RunMethod(
          config, request, global_params=global_params)

    Get.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations/{locationsId}',
        http_method='GET',
        method_id='apigateway.projects.locations.get',
        ordered_params=['name'],
        path_params=['name'],
        query_params=[],
        relative_path='v1beta/{+name}',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsGetRequest',
        response_type_name='ApigatewayLocation',
        supports_download=False,
    )

    def List(self, request, global_params=None):
      r"""Lists information about the supported locations for this service.

      Args:
        request: (ApigatewayProjectsLocationsListRequest) input message
        global_params: (StandardQueryParameters, default: None) global arguments
      Returns:
        (ApigatewayListLocationsResponse) The response message.
      """
      config = self.GetMethodConfig('List')
      return self._RunMethod(
          config, request, global_params=global_params)

    List.method_config = lambda: base_api.ApiMethodInfo(
        flat_path='v1beta/projects/{projectsId}/locations',
        http_method='GET',
        method_id='apigateway.projects.locations.list',
        ordered_params=['name'],
        path_params=['name'],
        query_params=['filter', 'pageSize', 'pageToken'],
        relative_path='v1beta/{+name}/locations',
        request_field='',
        request_type_name='ApigatewayProjectsLocationsListRequest',
        response_type_name='ApigatewayListLocationsResponse',
        supports_download=False,
    )

  class ProjectsService(base_api.BaseApiService):
    """Service class for the projects resource."""

    _NAME = 'projects'

    def __init__(self, client):
      super(ApigatewayV1beta.ProjectsService, self).__init__(client)
      self._upload_configs = {
          }



# -*- coding: utf-8 -*- #
# Copyright 2021 Google LLC. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Functionality related to Cloud Functions v2 API clients."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import re

from apitools.base.py import encoding
from apitools.base.py import exceptions as apitools_exceptions
import frozendict
from googlecloudsdk.api_lib.cloudresourcemanager import projects_api
from googlecloudsdk.api_lib.functions.v2 import exceptions
from googlecloudsdk.api_lib.util import apis
from googlecloudsdk.calliope import base as calliope_base
from googlecloudsdk.command_lib.projects import util as projects_util
from googlecloudsdk.core import log
from googlecloudsdk.core import properties
from googlecloudsdk.core.console import console_io
from googlecloudsdk.core.console import progress_tracker
from googlecloudsdk.core.util import encoding as encoder
from googlecloudsdk.core.util import retry
import six

_API_NAME = 'cloudfunctions'

_V2_ALPHA = 'v2alpha'
_V2_BETA = 'v2beta'
_V2_GA = 'v2'

RELEASE_TRACK_TO_API_VERSION = {
    calliope_base.ReleaseTrack.ALPHA: 'v2alpha',
    calliope_base.ReleaseTrack.BETA: 'v2beta',
    calliope_base.ReleaseTrack.GA: 'v2'
}

MAX_WAIT_MS = 1820000
SLEEP_MS = 1000

# EventArc types
EA_PUBSUB_MESSAGE_PUBLISHED = 'google.cloud.pubsub.topic.v1.messagePublished'
EA_STORAGE_ARCHIVE = 'google.cloud.storage.object.v1.archived'
EA_STORAGE_DELETE = 'google.cloud.storage.object.v1.deleted'
EA_STORAGE_FINALIZE = 'google.cloud.storage.object.v1.finalized'
EA_STORAGE_UPDATE = 'google.cloud.storage.object.v1.metadataUpdated'

EVENTARC_STORAGE_TYPES = (
    EA_STORAGE_ARCHIVE,
    EA_STORAGE_DELETE,
    EA_STORAGE_FINALIZE,
    EA_STORAGE_UPDATE,
)

# EventFlow types
EF_PUBSUB_MESSAGE_PUBLISH = 'google.pubsub.topic.publish'
EF_STORAGE_ARCHIVE = 'google.storage.object.archive'
EF_STORAGE_DELETE = 'google.storage.object.delete'
EF_STORAGE_FINALIZE = 'google.storage.object.finalize'
EF_STORAGE_METADATA_UPDATE = 'google.storage.object.metadataUpdate'

EVENTFLOW_TO_EVENTARC_STORAGE_MAP = frozendict.frozendict({
    EF_STORAGE_ARCHIVE: EA_STORAGE_ARCHIVE,
    EF_STORAGE_DELETE: EA_STORAGE_DELETE,
    EF_STORAGE_FINALIZE: EA_STORAGE_FINALIZE,
    EF_STORAGE_METADATA_UPDATE: EA_STORAGE_UPDATE,
})

# Legacy types
LEGACY_PUBSUB_MESSAGE_PUBLISH = (
    'providers/cloud.pubsub/eventTypes/topic.publish')

PUBSUB_MESSAGE_PUBLISH_TYPES = (
    EA_PUBSUB_MESSAGE_PUBLISHED,
    EF_PUBSUB_MESSAGE_PUBLISH,
    LEGACY_PUBSUB_MESSAGE_PUBLISH,
)


def GetProject():
  """Returns the value of the core/project config prooerty.

  Config properties can be overridden with command line flags. If the --project
  flag was provided, this will return the value provided with the flag.
  """
  return properties.VALUES.core.project.Get(required=True)


def GetMessagesModule(release_track):
  """Returns the API messages module for GCFv2."""
  api_version = RELEASE_TRACK_TO_API_VERSION.get(release_track)
  return apis.GetMessagesModule(_API_NAME, api_version)


def GetStage(messages):
  """Returns corresponding GoogleCloudFunctionsV2(alpha|beta|ga)Stage."""
  if messages is apis.GetMessagesModule(_API_NAME, _V2_ALPHA):
    return messages.GoogleCloudFunctionsV2alphaStage
  elif messages is apis.GetMessagesModule(_API_NAME, _V2_BETA):
    return messages.GoogleCloudFunctionsV2betaStage
  else:
    return messages.GoogleCloudFunctionsV2Stage


def GetStateMessage(messages):
  """Returns corresponding GoogleCloudFunctionsV2(alpha|beta|ga)stateMessage."""
  if messages is apis.GetMessagesModule(_API_NAME, _V2_ALPHA):
    return messages.GoogleCloudFunctionsV2alphaStateMessage
  elif messages is apis.GetMessagesModule(_API_NAME, _V2_BETA):
    return messages.GoogleCloudFunctionsV2betaStateMessage
  else:
    return messages.GoogleCloudFunctionsV2StateMessage


def GetClientInstance(release_track):
  """Returns an API client for GCFv2."""
  api_version = RELEASE_TRACK_TO_API_VERSION.get(release_track)
  return apis.GetClientInstance(_API_NAME, api_version)


def GetStateMessagesStrings(state_messages):
  """Returns the list of string representations of the state messages."""
  return map(lambda st: '[{}] {}'.format(str(st.severity), st.message),
             state_messages)


def _GetStageName(name_enum):
  """Converts NameValueValuesEnum into human-readable text."""
  return str(name_enum).replace('_', ' ').title()


def _BuildOperationMetadata(messages):
  """Returns corresponding GoogleCloudFunctionsV2(alpha|beta|ga)OperationMetadata."""
  if messages is apis.GetMessagesModule(_API_NAME, _V2_ALPHA):
    return messages.GoogleCloudFunctionsV2alphaOperationMetadata
  elif messages is apis.GetMessagesModule(_API_NAME, _V2_BETA):
    return messages.GoogleCloudFunctionsV2betaOperationMetadata
  elif messages is apis.GetMessagesModule(_API_NAME, _V2_GA):
    return messages.GoogleCloudFunctionsV2OperationMetadata
  else:
    raise NotImplementedError('Invalid messages module.')


def _GetOperationMetadata(messages, operation):
  return encoding.PyValueToMessage(
      _BuildOperationMetadata(messages),
      encoding.MessageToPyValue(operation.metadata))


def _GetOperation(client, request):
  """Get operation and return None if doesn't exist."""
  try:
    # We got response for a GET request, so an operation exists.
    return client.projects_locations_operations.Get(request)
  except apitools_exceptions.HttpError as error:
    if error.status_code == six.moves.http_client.NOT_FOUND:
      return None
    raise


def _GetStages(client, request, messages):
  """Returns None until stages have been loaded in the operation."""
  operation = _GetOperation(client, request)
  if operation.error:
    raise exceptions.StatusToFunctionsError(operation.error)

  if not operation.metadata:
    return None
  operation_metadata = _GetOperationMetadata(messages, operation)
  if not operation_metadata.stages:
    return None

  stages = []
  for stage in operation_metadata.stages:
    message = '[{}]'.format(_GetStageName(stage.name))
    stages.append(progress_tracker.Stage(message, key=str(stage.name)))
  return stages


def _GetOperationStatus(client, request, tracker, messages):
  """Returns a Boolean indicating whether the request has completed."""
  operation = client.projects_locations_operations.Get(request)
  if operation.error:
    raise exceptions.StatusToFunctionsError(
        operation.error, error_message=OperationErrorToString(operation.error))

  operation_metadata = _GetOperationMetadata(messages, operation)
  for stage in operation_metadata.stages:
    stage_in_progress = (
        stage.state is GetStage(messages).StateValueValuesEnum.IN_PROGRESS)
    stage_complete = (
        stage.state is GetStage(messages).StateValueValuesEnum.COMPLETE)

    if not stage_in_progress and not stage_complete:
      continue

    stage_key = str(stage.name)
    if tracker.IsComplete(stage_key):
      # Cannot update a completed stage in the tracker
      continue

    # Start running a stage
    if tracker.IsWaiting(stage_key):
      tracker.StartStage(stage_key)

    # Update stage message, including Build logs URL if applicable
    stage_message = stage.message or ''
    if stage_in_progress:
      stage_message = (stage_message or 'In progress') + '... '
    else:
      stage_message = ''

    if stage.resourceUri and stage_key == 'BUILD':
      stage_message += 'Logs are available at [{}]'.format(stage.resourceUri)

    tracker.UpdateStage(stage_key, stage_message)

    # Complete a finished stage
    if stage_complete:
      if stage.stateMessages:
        tracker.CompleteStageWithWarnings(
            stage_key, GetStateMessagesStrings(stage.stateMessages))
      else:
        tracker.CompleteStage(stage_key)
  return operation.done


def WaitForOperation(client,
                     messages,
                     operation,
                     description,
                     extra_stages=None):
  """Wait for a long-running operation (LRO) to complete.

  Args:
    client: The GCFv2 API client.
    messages: The GCFv2 message stubs.
    operation: The operation message response.
    description: str, the description of the waited operation.
    extra_stages: List[progress_tracker.Stage]|None, list of optional stages for
      the progress tracker to watch. The GCF 2nd api returns unexpected stages
      in the case of rollbacks.
  """
  request = messages.CloudfunctionsProjectsLocationsOperationsGetRequest(
      name=operation.name)
  # Wait for stages to be loaded.
  with progress_tracker.ProgressTracker('Preparing function') as tracker:
    retryer = retry.Retryer(max_wait_ms=MAX_WAIT_MS)
    try:
      # List[progress_tracker.Stage]
      stages = retryer.RetryOnResult(
          _GetStages,
          args=[client, request, messages],
          should_retry_if=None,
          sleep_ms=SLEEP_MS)
    except retry.WaitException:
      raise exceptions.FunctionsError('Operation {0} is taking too long'.format(
          request.name))

  if extra_stages is not None:
    stages += extra_stages

  # Wait for LRO to complete.
  description += '...'
  with progress_tracker.StagedProgressTracker(description, stages) as tracker:
    retryer = retry.Retryer(max_wait_ms=MAX_WAIT_MS)
    try:
      retryer.RetryOnResult(
          _GetOperationStatus,
          args=[client, request, tracker, messages],
          should_retry_if=False,
          sleep_ms=SLEEP_MS)
    except retry.WaitException:
      raise exceptions.FunctionsError('Operation {0} is taking too long'.format(
          request.name))


def FormatTimestamp(timestamp):
  """Formats a timestamp which will be presented to a user.

  Args:
    timestamp: Raw timestamp string in RFC3339 UTC "Zulu" format.

  Returns:
    Formatted timestamp string.
  """
  return re.sub(r'(\.\d{3})\d*Z$', r'\1', timestamp.replace('T', ' '))


def OperationErrorToString(error):
  """Returns a human readable string representation from the operation.

  Args:
    error: A string representing the raw json of the operation error.

  Returns:
    A human readable string representation of the error.
  """
  error_message = 'OperationError: code={0}, message={1}'.format(
      error.code, encoder.Decode(error.message))
  messages = apis.GetMessagesModule('cloudfunctions', _V2_ALPHA)
  if error.details:
    for detail in error.details:
      sub_error = encoding.PyValueToMessage(messages.Status,
                                            encoding.MessageToPyValue(detail))
      if sub_error.code is not None or sub_error.message is not None:
        error_message += '\n' + OperationErrorToString(sub_error)
  return error_message


def HasRoleBinding(sa_email, role):
  # type(str, str) -> bool
  """Returns whether the given service account has the given role bound.

  Args:
    sa_email: The service account to check.
    role: The role to check for.
  """
  iam_policy = projects_api.GetIamPolicy(
      projects_util.ParseProject(GetProject()))

  # iam_policy.bindings structure:
  # list[<Binding
  #       members=['serviceAccount:member@thing.iam.gserviceaccount.com', ...],
  #       role='roles/somerole'>...]
  return any(
      'serviceAccount:{}'.format(sa_email) in b.members and b.role == role
      for b in iam_policy.bindings)


def PromptToBindRoleIfMissing(sa_email, role, reason=''):
  # type: (str, str, str) -> None
  """Prompts to bind the role to the service account if missing.

  If the console cannot prompt, a warning is logged instead.

  Args:
    sa_email: The service account email to bind the role to.
    role: The role to bind if missing.
    reason: Extra information to print explaining why the binding is necessary.
  """
  if HasRoleBinding(sa_email, role):
    return

  log.status.Print('Service account [{}] is missing the role [{}].\n{}'.format(
      sa_email, role, reason))

  bind = console_io.CanPrompt() and console_io.PromptContinue(
      prompt_string='\nBind the role [{}] to service account [{}]?'.format(
          role, sa_email),
      cancel_on_no=False)
  if bind:
    project_ref = projects_util.ParseProject(GetProject())
    member = 'serviceAccount:{}'.format(sa_email)
    projects_api.AddIamPolicyBinding(project_ref, member, role)
    log.status.Print('Role successfully bound.\n')
  else:
    log.warning('Manual binding of above role may be necessary.\n')
